#! /usr/bin/env python
"""SearchSource: AIS electronic Library"""
from __future__ import annotations

import re
import urllib.parse
from pathlib import Path
from urllib.parse import urlparse

import requests
import zope.interface
from pydantic import Field

import colrev.exceptions as colrev_exceptions
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
import colrev.record.record_prep
from colrev.constants import Fields
from colrev.constants import FieldValues
from colrev.constants import SearchSourceHeuristicStatus
from colrev.constants import SearchType
from colrev.packages.ais_library.src import ais_load_utils

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.SearchSourceInterface)
class AISeLibrarySearchSource:
    """AIS electronic Library (AISeL)"""

    settings_class = colrev.package_manager.package_settings.DefaultSourceSettings

    # pylint: disable=colrev-missed-constant-usage
    source_identifier = "url"
    search_types = [
        SearchType.DB,
        SearchType.TOC,
        SearchType.API,
    ]
    endpoint = "colrev.ais_library"

    ci_supported: bool = Field(default=True)
    heuristic_status = SearchSourceHeuristicStatus.supported

    db_url = "https://aisel.aisnet.org/"

    _conference_abbreviations = {
        "ICIS": "International Conference on Information Systems",
        "PACIS": "Pacific-Asia Conference on Information Systems",
        "ECIS": "European Conference on Information Systems",
        "AMCIS": "Americas Conference on Information Systems",
        "HICSS": "Hawaii International Conference on System Sciences",
        "MCIS": "Mediterranean Conference on Information Systems",
        "ACIS": "Australasian Conference on Information Systems",
        "WHICEB": "Wuhan International Conference on e-Business",
        "CONF-IRM": "International Conference on Information Resources Management",
    }

    _link_confs = {
        "https://aisel.aisnet.org/hicss": "Hawaii International Conference on System Sciences",
        "https://aisel.aisnet.org/amcis": "Americas Conference on Information Systems",
        "https://aisel.aisnet.org/pacis": "Pacific-Asia Conference on Information Systems",
        "https://aisel.aisnet.org/ecis": "European Conference on Information Systems",
        "https://aisel.aisnet.org/icis": "International Conference on Information Systems",
    }

    def __init__(
        self, *, source_operation: colrev.process.operation.Operation, settings: dict
    ) -> None:
        self.search_source = self.settings_class(**settings)
        self.review_manager = source_operation.review_manager
        self.quality_model = self.review_manager.get_qm()
        self.source_operation = source_operation

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for AIS electronic Library (AISeL)"""

        result = {"confidence": 0.0}
        # TBD: aisel does not return bibtex?!
        nr_ais_links = data.count("https://aisel.aisnet.org/")
        nr_items = data.count("\n@")
        if nr_items > 0 and Fields.STATUS not in data:
            result["confidence"] = nr_ais_links / nr_items

        if "%T " in data and Fields.STATUS not in data:
            if data.count("%U https://aisel.aisnet.org") > 0.9 * data.count("%T "):
                result["confidence"] = 1.0

        if "@article" in data or "@inproc" in data and Fields.STATUS not in data:
            if data.count("https://aisel.aisnet.org") > 0.9 * data.count("\n@"):
                result["confidence"] = 1.0

        return result

    @classmethod
    def _parse_query(cls, *, query: str) -> dict:
        peer_reviewed = "peer_reviewed=true" in query
        start_date = ""
        if "start_date=" in query:
            start_date = query[query.find("start_date=") + 11 :]
            start_date = start_date[: start_date.find("&")]
            start_date = start_date.replace("%2F", "/")
        end_date = ""
        if "end_date=" in query:
            end_date = query[query.find("end_date=") + 9 :]
            end_date = end_date[: end_date.find("&")]
            end_date = end_date.replace("%2F", "/")

        query = query[query.find("?q=") + 3 : query.find("&start")]
        query_parts = urllib.parse.unquote(query).split(" ")

        search_terms = []
        query_parts_merged = []
        parenthesis_expression = ""
        for query_part in query_parts:
            if query_part not in ["(", ")"] and "" == parenthesis_expression:
                query_parts_merged.append(query_part)
            elif query_part == "(":
                parenthesis_expression += "("
            elif query_part == ")":
                parenthesis_expression = parenthesis_expression.rstrip().replace(
                    "(", ""
                )
                query_parts_merged.append(parenthesis_expression)
                parenthesis_expression = ""
            else:
                parenthesis_expression = parenthesis_expression + query_part + " "

        term_no = 1
        operator = ""

        for query_part in query_parts_merged:
            if query_part in ["OR", "AND", "NOT"]:
                operator = query_part
                continue

            field = "All fields"
            if "%3A" in query_part:
                field, query_part = query_part.split("%3A")
            search_term = {"operator": operator, "term": query_part, "field": field}

            search_terms.append(search_term)
            term_no += 1

        params = {"query": search_terms, "scope": {"peer_reviewed": peer_reviewed}}
        if start_date:
            params["start_date"] = start_date
        if end_date:
            params["end_date"] = end_date
        return params

    @classmethod
    def add_endpoint(
        cls,
        operation: colrev.ops.search.Search,
        params: str,
    ) -> colrev.settings.SearchSource:
        """Add SearchSource as an endpoint (based on query provided to colrev search --add )"""
        params_dict = {}
        if params:
            if params.startswith("http"):
                params_dict = {Fields.URL: params}
            else:
                for item in params.split(";"):
                    key, value = item.split("=")
                    params_dict[key] = value

        search_type = operation.select_search_type(
            search_types=cls.search_types, params=params_dict
        )

        if search_type == SearchType.DB:
            search_source = operation.create_db_source(
                search_source_cls=cls,
                params=params_dict,
            )

        # pylint: disable=colrev-missed-constant-usage
        elif search_type == SearchType.API:
            if "url" in params_dict:
                host = urlparse(params_dict["url"]).hostname
                assert host and host.endswith("aisel.aisnet.org")
                q_params = cls._parse_query(query=params_dict["url"])
                filename = operation.get_unique_filename(file_path_string="ais")
                search_source = colrev.settings.SearchSource(
                    endpoint=cls.endpoint,
                    filename=filename,
                    search_type=SearchType.API,
                    search_parameters=q_params,
                    comment="",
                )
            else:
                # Add API search without params
                search_source = operation.create_api_source(endpoint=cls.endpoint)

        # elif search_type == SearchType.TOC:
        else:
            raise NotImplementedError

        operation.add_source_and_search(search_source)
        return search_source

    def _validate_source(self) -> None:
        """Validate the SearchSource (parameters etc.)"""

        source = self.search_source

        self.review_manager.logger.debug(f"Validate SearchSource {source.filename}")

        if source.search_type == SearchType.API:
            if "query" not in source.search_parameters:
                # if "search_terms" not in source.search_parameters["query"]:
                raise colrev_exceptions.InvalidQueryException("query parameter missing")

        self.review_manager.logger.debug(f"SearchSource {source.filename} validated")

    def _get_ais_query_return(self) -> list:
        def query_from_params(params: dict) -> str:

            final = ""
            if isinstance(params["query"], str):
                final = params["query"]
            else:
                for i in params["query"]:
                    final = f"{final} {i['operator']} {i['field']}:{i['term']}".strip()
                    final = final.replace("All fields:", "")

            if "start_date" in params["query"]:
                final = f"{final}&start_date={params['query']['start_date']}"

            if "end_date" in params["query"]:
                final = f"{final}&end_date={params['query']['end_date']}"

            if "peer_reviewed" in params["query"]:
                final = f"{final}&peer_reviewed=true"

            return urllib.parse.quote(final)

        params = self.search_source.search_parameters
        final_q = query_from_params(params)

        query_string = (
            "https://aisel.aisnet.org/do/search/results/refer?"
            + "start=0&context=509156&sort=score&facet=&dlt=Export122204"
        )
        query_string = f"{query_string}&q={final_q}"

        response = requests.get(query_string, timeout=300)
        response.raise_for_status()

        # Note: the following writes the enl to the feed file (bib).
        # This file is replaced by ais_feed.save()

        records = colrev.loader.load_utils.loads(
            response.text,
            implementation="enl",
            id_labeler=ais_load_utils.enl_id_labeler,
            entrytype_setter=ais_load_utils.enl_entrytype_setter,
            field_mapper=ais_load_utils.enl_field_mapper,
            logger=self.review_manager.logger,
        )

        return list(records.values())

    def _run_api_search(
        self,
        *,
        ais_feed: colrev.ops.search_api_feed.SearchAPIFeed,
        rerun: bool,
    ) -> None:
        # pylint: disable=too-many-branches

        if rerun:
            self.review_manager.logger.info(
                "Performing a search of the full history (may take time)"
            )

        try:
            for record_dict in self._get_ais_query_return():
                try:
                    # Note : discard "empty" records
                    if "" == record_dict.get(
                        Fields.AUTHOR, ""
                    ) and "" == record_dict.get(Fields.TITLE, ""):
                        continue

                    prep_record = colrev.record.record_prep.PrepRecord(record_dict)
                    ais_feed.add_update_record(prep_record)

                except colrev_exceptions.NotFeedIdentifiableException:
                    continue
        except (
            requests.exceptions.JSONDecodeError,
            requests.exceptions.HTTPError,
        ) as exc:
            # watch github issue:
            # https://github.com/fabiobatalha/crossrefapi/issues/46
            if "504 Gateway Time-out" in str(exc):
                raise colrev_exceptions.ServiceNotAvailableException(
                    "Crossref (check https://status.crossref.org/)"
                )
            raise colrev_exceptions.ServiceNotAvailableException(
                f"Crossref (check https://status.crossref.org/) ({exc})"
            )

        ais_feed.save()

    def search(self, rerun: bool) -> None:
        """Run a search of AISeLibrary"""

        self._validate_source()

        if self.search_source.search_type == SearchType.API:
            ais_feed = self.search_source.get_api_feed(
                review_manager=self.review_manager,
                source_identifier=self.source_identifier,
                update_only=(not rerun),
            )
            self._run_api_search(
                ais_feed=ais_feed,
                rerun=rerun,
            )
        elif self.search_source.search_type == SearchType.DB:
            self.source_operation.run_db_search(  # type: ignore
                search_source_cls=self.__class__,
                source=self.search_source,
            )

    def prep_link_md(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.record.Record:
        """Not implemented"""
        return record

    def _load_enl(self) -> dict:

        return colrev.loader.load_utils.load(
            filename=self.search_source.filename,
            id_labeler=ais_load_utils.enl_id_labeler,
            entrytype_setter=ais_load_utils.enl_entrytype_setter,
            field_mapper=ais_load_utils.enl_field_mapper,
            logger=self.review_manager.logger,
        )

    def _load_bib(self) -> dict:

        records = colrev.loader.load_utils.load(
            filename=self.search_source.filename,
            logger=self.review_manager.logger,
            field_mapper=ais_load_utils.bib_field_mapper,
            unique_id_field="ID",
        )

        return records

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        # pylint: disable=colrev-missed-constant-usage
        if self.search_source.filename.suffix in [".txt", ".enl"]:
            return self._load_enl()

        # for API-based searches
        if self.search_source.filename.suffix == ".bib":
            return self._load_bib()

        raise NotImplementedError

    def _fix_entrytype(self, *, record: colrev.record.record.Record) -> None:
        # Note : simple heuristic
        # but at the moment, AISeLibrary only indexes articles and conference papers
        if (
            record.data.get(Fields.VOLUME, FieldValues.UNKNOWN) != FieldValues.UNKNOWN
            or record.data.get(Fields.NUMBER, FieldValues.UNKNOWN)
            != FieldValues.UNKNOWN
        ) and not any(
            x in record.data.get(Fields.JOURNAL, "")
            for x in [
                "HICSS",
                "ICIS",
                "ECIS",
                "AMCIS",
                "Proceedings",
                "All Sprouts Content",
            ]
        ):
            # Journal articles
            if (
                Fields.JOURNAL not in record.data
                and Fields.TITLE in record.data
                and Fields.CHAPTER in record.data
            ):
                record.rename_field(key=Fields.TITLE, new_key=Fields.JOURNAL)
                record.rename_field(key=Fields.CHAPTER, new_key=Fields.TITLE)
                record.remove_field(key=Fields.PUBLISHER)

            record.change_entrytype("article", qm=self.quality_model)
        else:
            # Inproceedings

            record.remove_field(key=Fields.PUBLISHER)

            if (
                Fields.BOOKTITLE not in record.data
                and Fields.TITLE in record.data
                and Fields.CHAPTER in record.data
            ):
                record.rename_field(key=Fields.TITLE, new_key=Fields.BOOKTITLE)
                record.rename_field(key=Fields.CHAPTER, new_key=Fields.TITLE)

            record.change_entrytype(
                new_entrytype="inproceedings", qm=self.quality_model
            )

            if record.data.get(Fields.BOOKTITLE, "") in [
                "Research-in-Progress Papers",
                "Research Papers",
            ]:
                if "https://aisel.aisnet.org/ecis" in record.data.get(Fields.URL, ""):
                    record.update_field(
                        key=Fields.BOOKTITLE, value="ECIS", source="prep_ais_source"
                    )

    def _unify_container_titles(self, *, record: colrev.record.record.Record) -> None:
        if "https://aisel.aisnet.org/misq/" in record.data.get(Fields.URL, ""):
            record.update_field(
                key=Fields.JOURNAL, value="MIS Quarterly", source="prep_ais_source"
            )
            record.remove_field(key=Fields.BOOKTITLE)

        if "https://aisel.aisnet.org/misqe/" in record.data.get(Fields.URL, ""):
            record.update_field(
                key=Fields.JOURNAL,
                value="MIS Quarterly Executive",
                source="prep_ais_source",
            )
            record.remove_field(key=Fields.BOOKTITLE)

        if "https://aisel.aisnet.org/bise/" in record.data.get(Fields.URL, ""):
            record.update_field(
                key=Fields.JOURNAL,
                value="Business & Information Systems Engineering",
                source="prep_ais_source",
            )
            record.remove_field(key=Fields.BOOKTITLE)

        if record.data[Fields.ENTRYTYPE] == "inproceedings":
            for conf_abbreviation, conf_name in self._conference_abbreviations.items():
                if conf_abbreviation in record.data.get(Fields.BOOKTITLE, ""):
                    record.update_field(
                        key=Fields.BOOKTITLE,
                        value=conf_name,
                        source="prep_ais_source",
                    )

        for link_part, conf_name in self._link_confs.items():
            if link_part in record.data.get(Fields.URL, ""):
                record.update_field(
                    key=Fields.BOOKTITLE,
                    value=conf_name,
                    source="prep_ais_source",
                )

    def _format_fields(self, *, record: colrev.record.record_prep.PrepRecord) -> None:
        if Fields.ABSTRACT in record.data:
            if record.data[Fields.ABSTRACT] == "N/A":
                record.remove_field(key=Fields.ABSTRACT)
        if Fields.AUTHOR in record.data:
            record.update_field(
                key=Fields.AUTHOR,
                value=record.data[Fields.AUTHOR].replace("\n", " "),
                source="prep_ais_source",
                keep_source_if_equal=True,
            )
        record.format_if_mostly_upper(Fields.TITLE, case="sentence")
        record.format_if_mostly_upper(Fields.JOURNAL, case=Fields.TITLE)
        record.format_if_mostly_upper(Fields.BOOKTITLE, case=Fields.TITLE)
        record.format_if_mostly_upper(Fields.AUTHOR, case=Fields.TITLE)

    def _exclude_complementary_material(
        self, *, record: colrev.record.record.Record
    ) -> None:
        if re.match(
            r"MISQ Volume \d{1,2}, Issue \d Table of Contents",
            record.data.get(Fields.TITLE, ""),
        ):
            record.prescreen_exclude(reason="complementary material")

    def prepare(
        self,
        record: colrev.record.record_prep.PrepRecord,
        source: colrev.settings.SearchSource,
    ) -> colrev.record.record.Record:
        """Source-specific preparation for the AIS electronic Library (AISeL)"""

        self._fix_entrytype(record=record)
        self._unify_container_titles(record=record)
        self._format_fields(record=record)
        self._exclude_complementary_material(record=record)
        return record



#!/usr/bin/env python3
"""The CoLRev review manager (main entrypoint)."""
from __future__ import annotations

from pathlib import Path


# pylint: disable=too-few-public-methods
# pylint: disable=too-many-instance-attributes
class PathManager:
    """Paths for the CoLRev review project (repository)"""

    SEARCH_DIR = Path("data/search")
    PREP_DIR = Path("data/prep")
    DEDUPE_DIR = Path("data/dedupe")
    PRESCREEN_DIR = Path("data/prescreen")
    PDF_DIR = Path("data/pdfs")
    SCREEN_DIR = Path("data/screen")
    DATA_DIR = Path("data/data")
    CORRECTIONS_DIR = Path(".corrections")
    OUTPUT_DIR = Path("output")
    RECORDS_FILE = Path("data/records.bib")
    SETTINGS_FILE = Path("settings.json")
    STATUS_FILE = Path("status.yaml")
    README_FILE = Path("readme.md")
    REPORT_FILE = Path(".report.log")
    GIT_IGNORE_FILE = Path(".gitignore")
    PRE_COMMIT_CONFIG = Path(".pre-commit-config.yaml")

    # Ensure the path uses forward slashes, which is compatible with Git's path handling
    RECORDS_FILE_GIT = str(RECORDS_FILE).replace("\\", "/")

    def __init__(self, base_path: Path) -> None:
        self.base_path = base_path

        self.search = base_path / self.SEARCH_DIR
        self.prep = base_path / self.PREP_DIR
        self.dedupe = base_path / self.DEDUPE_DIR
        self.prescreen = base_path / self.PRESCREEN_DIR
        self.pdf = base_path / self.PDF_DIR
        self.screen = base_path / self.SCREEN_DIR
        self.data = base_path / self.DATA_DIR
        self.corrections = base_path / self.CORRECTIONS_DIR
        self.output = base_path / self.OUTPUT_DIR
        self.records = base_path / self.RECORDS_FILE
        self.settings = base_path / self.SETTINGS_FILE
        self.status = base_path / self.STATUS_FILE
        self.readme = base_path / self.README_FILE
        self.report = base_path / self.REPORT_FILE
        self.git_ignore = base_path / self.GIT_IGNORE_FILE
        self.pre_commit_config = base_path / self.PRE_COMMIT_CONFIG



#! /usr/bin/env python
"""Functionality for individual records."""
from __future__ import annotations

import re
import typing

from nameparser import HumanName

import colrev.env.utils
import colrev.exceptions as colrev_exceptions
import colrev.record.record
from colrev.constants import Fields
from colrev.constants import FieldValues

if typing.TYPE_CHECKING:  # pragma: no cover
    import colrev.review_manager
    import colrev.record.qm.quality_model


NO_CAPS = ["of", "for", "the", "and"]
ALL_CAPS = ["IEEE", "ACM", "M&A", "B2B", "B2C", "C2C", "I"]
ALL_CAPS_DICT = {r"U\.S\.": "U.S."}


def capitalize_entities(input_str: str) -> str:
    """Utility function to capitalize entities"""

    for all_cap in ALL_CAPS:
        input_str = re.sub(
            rf"\b{all_cap.lower()}\b", all_cap.upper(), input_str, flags=re.IGNORECASE
        )

    for all_cap, repl in ALL_CAPS_DICT.items():
        input_str = re.sub(
            rf"\b{all_cap.lower()}\b", repl, input_str, flags=re.IGNORECASE
        )

    for no_cap in NO_CAPS:
        if input_str.lower().startswith(no_cap):
            continue
        input_str = re.sub(
            rf"\b{no_cap.lower()}\b", no_cap, input_str, flags=re.IGNORECASE
        )

    input_str = input_str.replace(" i'", " I'").replace("'S ", "'s ")

    input_str = re.sub(r"it-(\w)", r"IT-\1", input_str, flags=re.IGNORECASE)
    input_str = re.sub(r"is-(\w)", r"IS-\1", input_str, flags=re.IGNORECASE)

    return input_str


class PrepRecord(colrev.record.record.Record):
    """The PrepRecord class provides a range of convenience functions for record preparation"""

    @classmethod
    def format_author_field(cls, input_string: str) -> str:
        """Format the author field (recognizing first/last names based on HumanName parser)"""

        def mostly_upper_case(input_string: str) -> bool:
            input_string = input_string.replace(".", "").replace(",", "")
            words = input_string.split()
            if len(words) == 0:
                return False
            return sum(word.isupper() for word in words) / len(words) > 0.8

        input_string = input_string.replace("\n", " ")
        # DBLP appends identifiers to non-unique authors
        input_string = str(re.sub(r"[0-9]{4}", "", input_string))

        if " and " in input_string:
            names = input_string.split(" and ")
        elif input_string.count(";") > 1:
            names = input_string.split(";")
        elif input_string.count(",") > 1:
            names = input_string.split(" ")
        else:
            names = [input_string]
        author_string = ""
        for name in names:
            # Note: https://github.com/derek73/python-nameparser
            # is very effective (maybe not perfect)

            parsed_name = HumanName(name)
            if mostly_upper_case(input_string.replace(" and ", "").replace("Jr", "")):
                parsed_name.capitalize(force=True)

            # Fix typical parser error
            if parsed_name.last == "" and parsed_name.title != "":
                parsed_name.last = parsed_name.title

            # pylint: disable=chained-comparison
            # Fix: when first names are abbreviated, nameparser creates errors:
            if (
                len(parsed_name.last) <= 3
                and parsed_name.last.isupper()
                and len(parsed_name.first) > 3
                and not parsed_name.first.isupper()
            ):
                # in these casees, first and last names are confused
                author_name_string = parsed_name.first + ", " + parsed_name.last
            else:
                parsed_name.string_format = "{last} {suffix}, {first} {middle}"
                # '{last} {suffix}, {first} ({nickname}) {middle}'
                author_name_string = str(parsed_name).replace(" , ", ", ")
                # Note: there are errors for the following author:
                # JR Cromwell and HK Gardner
                # The JR is probably recognized as Junior.
                # Check whether this is fixed in the Grobid name parser

            if author_string == "":
                author_string = author_name_string
            else:
                author_string = author_string + " and " + author_name_string

        return author_string

    def format_if_mostly_upper(self, key: str, *, case: str = "sentence") -> None:
        """Format the field if it is mostly in upper case"""

        if key not in self.data or self.data[key] == FieldValues.UNKNOWN:
            return

        if colrev.env.utils.percent_upper_chars(self.data[key]) < 0.6:
            return

        # Note: the truecase package is not very reliable (yet)

        self.data[key] = self.data[key].replace("\n", " ")

        if case == "sentence":
            self.data[key] = self.data[key].capitalize()
        elif case == "title":
            self.data[key] = self.data[key].title()
        else:
            raise colrev_exceptions.ParameterError(
                parameter="case", value=case, options=["sentence", "title"]
            )

        self.data[key] = capitalize_entities(self.data[key])

    def unify_pages_field(self) -> None:
        """Unify the format of the page field"""
        if Fields.PAGES not in self.data:
            return
        if not isinstance(self.data[Fields.PAGES], str):
            return
        if 1 == self.data[Fields.PAGES].count("-"):
            self.data[Fields.PAGES] = self.data[Fields.PAGES].replace("-", "--")
        self.data[Fields.PAGES] = (
            self.data[Fields.PAGES]
            .replace("–", "--")
            .replace("----", "--")
            .replace(" -- ", "--")
            .rstrip(".")
        )
        if re.match(r"^\d+\-\-\d+$", self.data[Fields.PAGES]):
            from_page, to_page = re.findall(r"(\d+)", self.data[Fields.PAGES])
            if len(from_page) > len(to_page):
                self.data[Fields.PAGES] = (
                    f"{from_page}--{from_page[:-len(to_page)]}{to_page}"
                )



#! /usr/bin/env python
"""Discovering and using packages."""
from __future__ import annotations

import json
import os
import tempfile
import typing
from pathlib import Path

import inquirer
import requests
import toml
from bs4 import BeautifulSoup
from m2r import parse_from_file

import colrev.package_manager.colrev_internal_packages
from colrev.constants import EndpointType
from colrev.constants import Filepaths
from colrev.constants import SearchType


INTERNAL_PACKAGES = (
    colrev.package_manager.colrev_internal_packages.get_internal_packages_dict()
)


# pylint: disable=too-many-instance-attributes
class PackageDoc:
    """PackageDoc"""

    package_id: str
    version: str
    license: str
    authors: list
    documentation: str
    repository: str = ""

    # For monorepo packages
    package_dir: Path

    monorepo: bool

    package_metadata: dict

    description: str
    dev_status: str
    endpoints: list
    search_types: list

    docs_package_readme_path: Path
    docs_rst_path: Path

    def __init__(self, package_id: str) -> None:
        self.package_id = package_id

        methods = [
            self._initialize_from_colrev_monorepo,
            self._initialize_from_pypi,
        ]

        for method in methods:
            if method(package_id):
                break
        else:
            raise NotImplementedError(
                f"Package {package_id} not found on PyPI/in CoLRev monorepo"
            )

        main_section = self.package_metadata["tool"]["poetry"]
        self.license = main_section["license"]
        self.version = main_section["version"]
        self.authors = main_section["authors"]
        self.documentation = main_section.get("documentation", None)
        self.repository = main_section.get("repository", None)
        self.endpoints = list(
            main_section.get("plugins", {}).get("colrev", {"na": "na"}).keys()
        )

        colrev_section = self.package_metadata.get("tool", {}).get("colrev", {})
        self.description = colrev_section.get("colrev_doc_description", "NA")
        self.search_types = colrev_section.get("search_types", [])

        colrev_doc_link = colrev_section.get("colrev_doc_link")
        self._set_docs_package_readme_path(colrev_doc_link)

        self.docs_rst_path = Path(f"{self.package_id}.rst")

    def _initialize_from_colrev_monorepo(self, package_id: str) -> bool:
        self.monorepo = True

        if package_id in INTERNAL_PACKAGES:
            self.package_dir = Path(INTERNAL_PACKAGES[package_id])
            with open(
                self.package_dir / Path("pyproject.toml"), encoding="utf-8"
            ) as file:
                self.package_metadata = toml.load(file)

            assert str(self.package_dir).endswith(
                package_id.replace("colrev.", "")
            ), package_id

            return True
        return False

    def _initialize_from_pypi(self, package_id: str) -> bool:
        self.monorepo = False

        response = requests.get(f"https://pypi.org/pypi/{package_id}/json", timeout=30)
        if response.status_code != 200:
            return False
        try:
            repo_url = response.json()["info"].get("project_urls", {}).get("Repository")
        except AttributeError:
            print(f"Failed to get repository URL for package {package_id}")
            return False
        gh_response = requests.get(f"{repo_url}/raw/main/pyproject.toml", timeout=30)
        if response.status_code != 200:
            return False

        self.package_metadata = toml.loads(gh_response.text)

        return True

    def has_endpoint(self, endpoint_type: EndpointType) -> bool:
        """Check if the package has a specific endpoint type"""

        return endpoint_type.value in self.endpoints

    def _get_authors_for_docs(self) -> str:
        """Get the authors for the documentation  (without emails in <>)"""

        authors = []
        for author in self.authors:
            authors.append(author.split("<")[0].strip())

        return ", ".join(authors)

    def _get_docs_short_description(self) -> str:
        return (
            self.description
            + " (:doc:`instructions </manual/packages/"
            + f"{self.package_id}>`)"
        )

    def _set_docs_package_readme_path(self, colrev_doc_link: str) -> None:

        if self.monorepo:
            self.docs_package_readme_path = self.package_dir / colrev_doc_link
        else:
            response = requests.get(
                self.repository + "/raw/main/" + colrev_doc_link, timeout=30
            )
            if response.status_code != 200:
                raise ValueError("Failed to download package readme from repository")
            with tempfile.NamedTemporaryFile(delete=False) as temp_file:
                temp_file.write(response.content)
                self.docs_package_readme_path = Path(temp_file.name)

        readme_content = self.docs_package_readme_path.read_text(encoding="utf-8")
        if not readme_content.startswith("## Summary"):
            raise ValueError(
                f"Package {self.package_id} readme does not start with '## Summary'"
            )

    # pylint: disable=line-too-long
    # pylint: disable=too-many-branches
    # pylint: disable=too-many-statements
    # flake8: noqa: E501
    def _get_header_info(self) -> str:

        # To format the table (adjust row height), the following is suggested:
        # from bs4 import BeautifulSoup

        # # Parse the generated HTML
        # with open('output.html', 'r') as f:
        #     soup = BeautifulSoup(f, 'html.parser')

        # # Find the table and add the ID
        # table = soup.find('table')
        # table['id'] = 'endpoint_overview_container'

        # # Write the modified HTML back to the file
        # with open('output.html', 'w') as f:
        #     f.write(str(soup))
        header_info = ""
        header_info += ".. |EXPERIMENTAL| image:: https://img.shields.io/badge/status-experimental-blue\n"
        header_info += "   :height: 14pt\n"
        header_info += "   :target: https://colrev-environment.github.io/colrev/dev_docs/dev_status.html\n"
        header_info += ".. |MATURING| image:: https://img.shields.io/badge/status-maturing-yellowgreen\n"
        header_info += "   :height: 14pt\n"
        header_info += "   :target: https://colrev-environment.github.io/colrev/dev_docs/dev_status.html\n"
        header_info += ".. |STABLE| image:: https://img.shields.io/badge/status-stable-brightgreen\n"
        header_info += "   :height: 14pt\n"
        header_info += "   :target: https://colrev-environment.github.io/colrev/dev_docs/dev_status.html\n"
        header_info += ".. |VERSION| image:: /_static/svg/iconmonstr-product-10.svg\n"
        header_info += "   :width: 15\n"
        header_info += "   :alt: Version\n"
        header_info += ".. |GIT_REPO| image:: /_static/svg/iconmonstr-code-fork-1.svg\n"
        header_info += "   :width: 15\n"
        header_info += "   :alt: Git repository\n"
        header_info += ".. |LICENSE| image:: /_static/svg/iconmonstr-copyright-2.svg\n"
        header_info += "   :width: 15\n"
        header_info += "   :alt: Licencse\n"
        header_info += ".. |MAINTAINER| image:: /_static/svg/iconmonstr-user-29.svg\n"
        header_info += "   :width: 20\n"
        header_info += "   :alt: Maintainer\n"
        header_info += (
            ".. |DOCUMENTATION| image:: /_static/svg/iconmonstr-book-17.svg\n"
        )
        header_info += "   :width: 15\n"
        header_info += "   :alt: Documentation\n"

        header_info += f"{self.package_id}\n"
        header_info += "=" * len(self.package_id) + "\n\n"
        header_info += f"|VERSION| Version: {self.version}\n\n"
        header_info += f"|MAINTAINER| Maintainer: {self._get_authors_for_docs()}\n\n"
        header_info += f"|LICENSE| License: {self.license}  \n\n"
        if self.repository != "":
            repo_name = self.repository.replace("https://github.com/", "")
            if "CoLRev-Environment/colrev" in repo_name:
                repo_name = "CoLRev-Environment/colrev"
            header_info += (
                f"|GIT_REPO| Repository: `{repo_name} <{self.repository}>`_ \n\n"
            )

        if self.documentation:
            header_info += (
                f"|DOCUMENTATION| `External documentation <{self.documentation}>`_\n\n"
            )

        header_info += ".. list-table::\n"
        header_info += "   :header-rows: 1\n"
        header_info += "   :widths: 20 30 80\n\n"
        header_info += "   * - Endpoint\n"
        header_info += "     - Status\n"
        header_info += "     - Add\n"

        for endpoint_type in EndpointType:
            if self.has_endpoint(endpoint_type):
                header_info += f"   * - {endpoint_type.value}\n"
                header_info += f"     - |{self.dev_status.upper()}|\n"
                if endpoint_type == EndpointType.review_type:
                    header_info += f"     - .. code-block:: \n\n\n         colrev init --type {self.package_id}\n\n"
                elif endpoint_type == EndpointType.search_source:
                    header_info += f"     - .. code-block:: \n\n\n         colrev search --add {self.package_id}\n\n"
                elif endpoint_type == EndpointType.prep:
                    header_info += f"     - .. code-block:: \n\n\n         colrev prep --add {self.package_id}\n\n"
                elif endpoint_type == EndpointType.prep_man:
                    header_info += f"     - .. code-block:: \n\n\n         colrev prep-man --add {self.package_id}\n\n"
                elif endpoint_type == EndpointType.dedupe:
                    header_info += f"     - .. code-block:: \n\n\n         colrev dedupe --add {self.package_id}\n\n"
                elif endpoint_type == EndpointType.prescreen:
                    header_info += f"     - .. code-block:: \n\n\n         colrev prescreen --add {self.package_id}\n\n"
                elif endpoint_type == EndpointType.pdf_get:
                    header_info += f"     - .. code-block:: \n\n\n         colrev pdf-get --add {self.package_id}\n\n"
                elif endpoint_type == EndpointType.pdf_get_man:
                    header_info += f"     - .. code-block:: \n\n\n         colrev pdf-get-man --add {self.package_id}\n\n"
                elif endpoint_type == EndpointType.pdf_prep:
                    header_info += f"     - .. code-block:: \n\n\n         colrev pdf-prep --add {self.package_id}\n\n"
                elif endpoint_type == EndpointType.pdf_prep_man:
                    header_info += f"     - .. code-block:: \n\n\n         colrev pdf-prep-man --add {self.package_id}\n\n"
                elif endpoint_type == EndpointType.screen:
                    header_info += f"     - .. code-block:: \n\n\n         colrev screen --add {self.package_id}\n\n"
                elif endpoint_type == EndpointType.data:
                    header_info += f"     - .. code-block:: \n\n\n         colrev data --add {self.package_id}\n\n"

        return header_info

    def import_package_docs(self) -> None:
        """Import the package documentation"""

        with open(
            Filepaths.COLREV_PATH
            / Path(f"docs/source/manual/packages/{self.docs_rst_path}"),
            "w",
            encoding="utf-8",
        ) as file:
            file.write(self._get_header_info())
            output = parse_from_file(self.docs_package_readme_path)
            output = output.replace(
                ".. list-table::", ".. list-table::\n   :align: left"
            )
            file.write(output)

    def get_endpoint_item(self, endpoint_type: EndpointType) -> dict:
        """Get the endpoint item for the package

        Format:
        {
            "package_endpoint_identifier": package_id,
            "status": status,
            "short_description": docs_short_description,
            "search_types": search_types,
        }
        """

        status = (
            self.dev_status.replace("stable", "|STABLE|")
            .replace("maturing", "|MATURING|")
            .replace("experimental", "|EXPERIMENTAL|")
        )
        endpoint_item = {
            "package_endpoint_identifier": self.package_id,
            "status": status,
            "short_description": self._get_docs_short_description(),
        }

        if endpoint_type == EndpointType.search_source:
            endpoint_item["search_types"] = self.search_types  # type: ignore

        return endpoint_item

    def get_docs_item(self) -> dict:
        """Get the documentation item for the package

        Format:
        {
            "identifier": package_id,
            "short_description": short_description,
            "path": docs_rst_path,
        }

        """
        item = {
            "identifier": self.package_id,
            "short_description": self._get_docs_short_description(),
            "path": self.docs_rst_path,
        }
        return item

    def __repr__(self) -> str:
        package_str = f"Package name: {self.package_id}\n"
        package_str += f"Package license: {self.license}\n"
        package_str += f"Package endpoints: {self.endpoints}\n"
        package_str += f"Package search types: {self.search_types}\n"
        package_str += f"Package authors: {self.authors}\n"
        package_str += f"Package documentation: {self.docs_package_readme_path}\n"
        package_str += f"Package description: {self.description}\n"

        return package_str


# pylint: disable=too-few-public-methods
class DocRegistryManager:
    """DocRegistryManager"""

    # Overview page of packages: rst
    docs_packages_index_path = Filepaths.COLREV_PATH / Path(
        "docs/source/manual/packages.rst"
    )
    # Overview page of packages: json
    docs_packages_overview_json_file = Filepaths.COLREV_PATH / Path(
        "docs/source/manual/packages_overview.json"
    )

    # Overviews of endpoints
    docs_package_endpoints_json_file = Filepaths.COLREV_PATH / Path(
        "docs/source/manual/package_endpoints.json"
    )

    # Overviews of search source types
    docs_search_source_types_json_file = Filepaths.COLREV_PATH / Path(
        "docs/source/manual/search_source_types.json"
    )

    pypi_ignored_packages_file = Filepaths.COLREV_PATH / Path(
        "docs/source/manual/pypi_ignored_packages.json"
    )

    package_endpoints_json: typing.Dict[str, list] = {x.value: [] for x in EndpointType}
    docs_for_index: typing.Dict[str, list] = {x.value: [] for x in EndpointType}
    packages_data: typing.Dict[str, dict] = {}

    def __init__(self) -> None:
        self.packages: typing.List[PackageDoc] = []

        with open(Filepaths.PACKAGES_JSON, encoding="utf-8") as file:
            self.packages_data = json.load(file)

    def _load_packages(self) -> None:
        for package_id, package_data in self.packages_data.items():
            try:
                package_doc = PackageDoc(package_id)
                package_doc.dev_status = package_data["dev_status"]
                self.packages.append(package_doc)
            except (
                toml.decoder.TomlDecodeError,
                NotImplementedError,
                ValueError,
            ) as exc:
                print(exc)
                print(f"Error loading package {package_id}")

        # Add package endpoints
        os.chdir(Filepaths.COLREV_PATH)
        for package in self.packages:
            for endpoint_type in EndpointType:
                if not package.has_endpoint(endpoint_type):
                    continue

                print(f"-  {package.package_id} / {endpoint_type.value}")

                package.import_package_docs()

                self.package_endpoints_json[endpoint_type.value].append(
                    package.get_endpoint_item(endpoint_type)
                )
                self.docs_for_index[endpoint_type.value].append(package.get_docs_item())

    def _extract_search_source_types(self) -> None:
        search_source_types: typing.Dict[str, list] = {}
        for search_source_type in SearchType:
            if search_source_type.value not in search_source_types:
                search_source_types[search_source_type.value] = []
            for search_source in self.package_endpoints_json["search_source"]:
                if search_source_type.value in search_source["search_types"]:
                    search_source_types[search_source_type.value].append(search_source)

        for key in search_source_types:
            search_source_types[key] = sorted(
                search_source_types[key],
                key=lambda d: d["package_endpoint_identifier"],
            )

        json_object = json.dumps(search_source_types, indent=4)
        with open(
            self.docs_search_source_types_json_file, "w", encoding="utf-8"
        ) as file:
            file.write(json_object)
            file.write("\n")  # to avoid pre-commit/eof-fix changes

    def _update_package_endpoints_json(self) -> None:
        for key in self.package_endpoints_json.keys():
            self.package_endpoints_json[key] = sorted(
                self.package_endpoints_json[key],
                key=lambda d: d["package_endpoint_identifier"],
            )

        self.docs_package_endpoints_json_file.unlink(missing_ok=True)
        json_object = json.dumps(self.package_endpoints_json, indent=4)
        with open(self.docs_package_endpoints_json_file, "w", encoding="utf-8") as file:
            file.write(json_object)
            file.write("\n")  # to avoid pre-commit/eof-fix changes

        with open(Filepaths.PACKAGES_ENDPOINTS_JSON, "w", encoding="utf-8") as file:
            file.write(json_object)
            file.write("\n")  # to avoid pre-commit/eof-fix changes

    def _update_packages_overview(self) -> None:
        packages_overview = []
        # for key, packages in self.package_endpoints_json.items():

        for endpoint_type in [x.value for x in EndpointType]:
            packages = self.package_endpoints_json[endpoint_type]
            for package in packages:
                package["endpoint_type"] = endpoint_type
                packages_overview.append(package)

        self.docs_packages_overview_json_file.unlink(missing_ok=True)
        json_object = json.dumps(packages_overview, indent=4)
        with open(self.docs_packages_overview_json_file, "w", encoding="utf-8") as file:
            file.write(json_object)
            file.write("\n")  # to avoid pre-commit/eof-fix changes

    def _write_docs_for_index(self) -> None:
        """Writes data from self.docs_for_index to the packages.rst file."""

        docs_packages_index_path_content = self.docs_packages_index_path.read_text(
            encoding="utf-8"
        )
        new_doc = []
        # append header
        for line in docs_packages_index_path_content.split("\n"):
            new_doc.append(line)
            if ":caption:" in line:
                new_doc.append("")
                break

        # append new links
        for endpoint_type in [x.value for x in EndpointType]:
            new_doc.append("")
            new_doc.append("")

            doc_items = self.docs_for_index[endpoint_type]
            for doc_item in sorted(doc_items, key=lambda d: d["identifier"]):
                if doc_item == "NotImplemented":
                    print(doc_item["path"])
                    continue
                new_doc.append(f"   packages/{doc_item['path']}")

        with open(self.docs_packages_index_path, "w", encoding="utf-8") as file:
            for line in new_doc:
                file.write(line + "\n")

    def _update_pypi_packages(self) -> None:
        """Retrieves CoLRev packages from PyPI"""

        def ask_package_action(package_name: str) -> str:
            questions = [
                inquirer.List(
                    "action",
                    message=f"What should be done with the package '{package_name}'?",
                    choices=["Ignore", "Add to colrev packages/docs"],
                )
            ]
            answer = inquirer.prompt(questions)
            return answer["action"]

        response = requests.get("https://pypi.org/search/?q=colrev", timeout=30)

        # read self.pypi_ignored_packages_file
        with open(self.pypi_ignored_packages_file, encoding="utf-8") as file:
            pypi_ignored_packages = json.load(file)

        if response.status_code == 200:

            soup = BeautifulSoup(response.text, "html.parser")
            package_snippets = soup.find_all("a", class_="package-snippet")
            search_results = []
            for package_snippet in package_snippets:
                span_elems = package_snippet.find_all("span")
                name = span_elems[0].text.strip()
                if "colrev" not in name:
                    continue
                search_results.append(
                    {
                        "name": name,
                        "version": span_elems[1].text.strip(),
                        "release_date": span_elems[2].text.strip(),
                        "description": package_snippet.p.text.strip(),
                    }
                )
            for search_result in search_results:
                package_name = search_result["name"]
                if package_name in pypi_ignored_packages:
                    continue
                if package_name in self.packages_data:
                    continue

                action = ask_package_action(package_name)

                if action == "Ignore":
                    print(f"The package '{package_name}' will be ignored.")
                    pypi_ignored_packages[package_name] = {}

                elif action == "Add to colrev packages/docs":
                    print(
                        f"The package '{package_name}' will be added to colrev packages/docs."
                    )
                    self.packages_data[package_name] = {"dev_status": "experimental"}

            with open(self.pypi_ignored_packages_file, "w", encoding="utf-8") as file:
                json.dump(pypi_ignored_packages, file, indent=4)
            with open(Filepaths.PACKAGES_JSON, "w", encoding="utf-8") as file:
                json.dump(self.packages_data, file, indent=4)

    def update(self) -> None:
        """Update the package endpoints and the package status."""

        self._update_pypi_packages()
        self._load_packages()

        self._update_package_endpoints_json()
        self._extract_search_source_types()
        self._update_packages_overview()
        self._write_docs_for_index()



#! /usr/bin/env python
"""Checker for erroneous-title-field."""
from __future__ import annotations

import colrev.record.qm.quality_model
from colrev.constants import DefectCodes
from colrev.constants import Fields

# pylint: disable=too-few-public-methods


class ErroneousTitleFieldChecker:
    """The ErroneousTitleFieldChecker"""

    msg = DefectCodes.ERRONEOUS_TITLE_FIELD

    def __init__(
        self, quality_model: colrev.record.qm.quality_model.QualityModel
    ) -> None:
        self.quality_model = quality_model

    def run(self, *, record: colrev.record.record.Record) -> None:
        """Run the erroneous-title-field checks"""

        if Fields.TITLE not in record.data or record.ignored_defect(
            key=Fields.TITLE, defect=self.msg
        ):
            return
        if record.masterdata_is_curated():
            return

        if self._title_has_errors(title=record.data[Fields.TITLE]):
            record.add_field_provenance_note(key=Fields.TITLE, note=self.msg)

        else:
            record.remove_field_provenance_note(key=Fields.TITLE, note=self.msg)

    def _title_has_errors(self, *, title: str) -> bool:
        # Cover common errors
        if title in {
            "A I S ssociation for nformation ystems",
            "The International Journal of Information Systems "
            + "Applications Chairman of the Editorial Board",
        }:
            return True

        if sum(char.isdigit() for char in title) > sum(
            char.isalpha() for char in title
        ):
            return True

        if " " not in title and (
            any(x in title for x in ["_", "."]) or any(char.isdigit() for char in title)
        ):
            return True
        return False


def register(quality_model: colrev.record.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(ErroneousTitleFieldChecker(quality_model))



#! /usr/bin/env python
"""Custom script setup."""
import typing
from pathlib import Path

import colrev.env.utils
import colrev.settings
from colrev.constants import SearchType

if typing.TYPE_CHECKING:  # pragma: no cover
    import colrev.review_manager


def setup_custom_search_script(
    *, review_manager: colrev.review_manager.ReviewManager
) -> None:
    """Setup a custom search script"""

    filedata = colrev.env.utils.get_package_file_content(
        module="colrev.ops",
        filename=Path("custom_scripts/custom_search_source_script.py"),
    )

    if filedata:
        with open("custom_search_source_script.py", "w", encoding="utf-8") as file:
            file.write(filedata.decode("utf-8"))

    review_manager.dataset.add_changes(Path("custom_search_source_script.py"))

    new_source = colrev.settings.SearchSource(
        endpoint="custom_search_source_script",
        filename=Path("data/search/custom_search.bib"),
        search_type=SearchType.DB,
        search_parameters={},
        comment="",
    )

    review_manager.settings.sources.append(new_source)
    review_manager.save_settings()



#! /usr/bin/env python
"""Template for a custom Prescreen PackageEndpoint"""
from __future__ import annotations

import random

import zope.interface

import colrev.package_manager.interfaces
import colrev.package_manager.package_settings
import colrev.process.operation
import colrev.record.record


# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.package_manager.interfaces.PrescreenInterface)
class CustomPrescreen:
    """Class for custom prescreen scripts"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings

    def __init__(
        self,
        *,
        prescreen_operation: colrev.ops.prescreen.Prescreen,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = self.settings_class(**settings)

    def run_prescreen(
        self,
        prescreen_operation: colrev.ops.prescreen.Prescreen,
        records: dict,
        split: list,  # pylint: disable=unused-argument
    ) -> dict:
        """Prescreen the record"""

        for record in records.values():
            if random.random() < 0.5:  # nosec
                prescreen_operation.prescreen(
                    record=colrev.record.record.Record(record), prescreen_inclusion=True
                )

            else:
                prescreen_operation.prescreen(
                    record=colrev.record.record.Record(record),
                    prescreen_inclusion=False,
                )

        prescreen_operation.review_manager.dataset.save_records_dict(records)
        prescreen_operation.review_manager.dataset.create_commit(
            msg="Pre-screen (random)",
            manual_author=False,
            script_call="colrev prescreen",
        )

        return records



#! /usr/bin/env python
"""Export of references in different bibliographical formats as a data operation"""
from __future__ import annotations

import copy
from enum import Enum
from pathlib import Path

import inquirer
import zope.interface
from pydantic import BaseModel
from pydantic import Field

import colrev.env.docker_manager
import colrev.env.utils
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Fields
from colrev.constants import FieldSet
from colrev.constants import RecordState
from colrev.writer.write_utils import write_file


class BibFormats(Enum):
    """Enum of available bibliography formats"""

    # pylint: disable=invalid-name
    zotero = "zotero"
    jabref = "jabref"
    citavi = "citavi"
    BiBTeX = "BiBTeX"
    RIS = "RIS"
    CSV = "CSV"
    EXCEL = "EXCEL"
    # endnote = "endnote"
    # mendeley = "mendeley"
    # rdf_bibliontology = "rdf_bibliontology"


class BibliographyExportSettings(
    colrev.package_manager.package_settings.DefaultSettings, BaseModel
):
    """Settings for BibliographyExport"""

    endpoint: str
    version: str
    bib_format: BibFormats


@zope.interface.implementer(colrev.package_manager.interfaces.DataInterface)
class BibliographyExport:
    """Export the sample references in Endpoint format"""

    settings: BibliographyExportSettings
    settings_class = BibliographyExportSettings

    ci_supported: bool = Field(default=True)

    # A challenge for the incremental mode is that data is run every time
    # the status runs (potentially creating very small increments)

    def __init__(
        self,
        *,
        data_operation: colrev.ops.data.Data,
        settings: dict,
    ) -> None:
        self.review_manager = data_operation.review_manager

        if "bib_format" not in settings:
            settings["bib_format"] = "endnote"
        settings["bib_format"] = BibFormats[settings["bib_format"]]
        if "version" not in settings:
            settings["version"] = "0.1"

        self.settings = self.settings_class(**settings)
        self.endpoint_path = self.review_manager.paths.output

    def _export(self, *, selected_records: dict) -> None:
        self.review_manager.logger.info(f"Export {self.settings.bib_format.name}")

        if self.settings.bib_format is BibFormats.zotero:
            export_filepath = self.endpoint_path / Path("zotero.bib")

        elif self.settings.bib_format is BibFormats.jabref:
            export_filepath = self.endpoint_path / Path("jabref.bib")

        elif self.settings.bib_format is BibFormats.citavi:
            export_filepath = self.endpoint_path / Path("citavi.bib")

        elif self.settings.bib_format is BibFormats.BiBTeX:
            export_filepath = self.endpoint_path / Path("references.bib")

        elif self.settings.bib_format is BibFormats.RIS:
            export_filepath = self.endpoint_path / Path("references.ris")

        elif self.settings.bib_format is BibFormats.CSV:
            export_filepath = self.endpoint_path / Path("references.csv")

        elif self.settings.bib_format is BibFormats.EXCEL:
            export_filepath = self.endpoint_path / Path("references.xlsx")
        else:
            raise NotImplementedError

        write_file(records_dict=selected_records, filename=export_filepath)

        self.review_manager.dataset.add_changes(export_filepath)
        self.review_manager.dataset.create_commit(
            msg=f"Create {self.settings.bib_format.name} bibliography",
        )

    @classmethod
    def add_endpoint(cls, operation: colrev.ops.data.Data, params: str) -> None:
        """Add bibliography as an endpoint"""

        add_package = {
            "endpoint": "colrev.bibliography_export",
            "version": "0.1",
            "bib_format": "endnote",
        }

        if params:
            assert params in [b.value for b in BibFormats]
            add_package["bib_format"] = params
        else:
            questions = [
                inquirer.List(
                    "bib_format",
                    message="Select a bibliography format",
                    choices=[b.value for b in BibFormats],
                ),
            ]
            choice = inquirer.prompt(questions)["bib_format"]
            add_package["bib_format"] = choice

        operation.review_manager.settings.data.data_package_endpoints.append(
            add_package
        )
        operation.review_manager.save_settings()
        operation.review_manager.dataset.create_commit(
            msg=f"Add colrev.bibliography_export: {add_package['bib_format']}"
        )

        instance = cls(data_operation=operation, settings=add_package)
        records = operation.review_manager.dataset.load_records_dict()
        instance.update_data(records, {}, silent_mode=True)

    # pylint: disable=unused-argument
    def update_data(
        self,
        records: dict,
        synthesized_record_status_matrix: dict,
        silent_mode: bool,
    ) -> None:
        """Update the data/bibliography"""

        self.endpoint_path.mkdir(exist_ok=True, parents=True)

        selected_records_original = {
            ID: r
            for ID, r in records.items()
            if r[Fields.STATUS]
            in [
                RecordState.rev_included,
                RecordState.rev_synthesized,
            ]
        }
        selected_records = copy.deepcopy(selected_records_original)
        for record in selected_records.values():
            for key_candidate in list(record.keys()):
                if key_candidate not in FieldSet.IDENTIFYING_FIELD_KEYS + [
                    Fields.ENTRYTYPE,
                    Fields.ID,
                    Fields.FILE,
                    "link",
                    Fields.URL,
                ]:
                    del record[key_candidate]
            # TBD: maybe resolve file paths (symlinks to absolute paths)?

        try:

            self._export(selected_records=selected_records)

        except NotImplementedError:
            self.review_manager.logger.info(
                f"Not yet implemented ({self.settings.bib_format})"
            )

    def update_record_status_matrix(
        self,
        synthesized_record_status_matrix: dict,
        endpoint_identifier: str,
    ) -> None:
        """Update the record_status_matrix"""
        # Note : automatically set all to True / synthesized
        for syn_id in list(synthesized_record_status_matrix.keys()):
            synthesized_record_status_matrix[syn_id][endpoint_identifier] = True

    def get_advice(
        self,
    ) -> dict:
        """Get advice on the next steps (for display in the colrev status)"""

        data_endpoint = "Data operation [bibliography export data endpoint]: "

        advice = {
            "msg": f"{data_endpoint}"
            + f"\n    - The references are updated automatically ({self.endpoint_path})",
            "detailed_msg": "TODO",
        }
        return advice



#! /usr/bin/env python
"""Utils for Unpaywall"""
from colrev.env.environment_manager import EnvironmentManager


UNPAYWALL_EMAIL_PATH = "packages.pdf_get.colrev.unpaywall.email"


def get_email() -> str:
    """Get user's name and email,

    if user have specified an email in registry, that will be returned
    otherwise it will return the email used in git
    """

    env_man = EnvironmentManager()
    env_mail = env_man.get_settings_by_key(UNPAYWALL_EMAIL_PATH)
    _, email = env_man.get_name_mail_from_git()
    email = env_mail or email
    return email



#! /usr/bin/env python
"""Conslidation of metadata based on LocalIndex as a prep operation"""
from __future__ import annotations

import zope.interface
from pydantic import Field

import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.packages.local_index.src.local_index as local_index_connector
import colrev.record.record
from colrev.constants import Fields

# pylint: disable=duplicate-code


# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.package_manager.interfaces.PrepInterface)
class LocalIndexPrep:
    """Prepares records based on LocalIndex metadata"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings

    ci_supported: bool = Field(default=True)

    source_correction_hint = (
        "correct the metadata in the source "
        + "repository (as linked in the provenance field)"
    )
    always_apply_changes = True

    def __init__(self, *, prep_operation: colrev.ops.prep.Prep, settings: dict) -> None:
        self.settings = self.settings_class(**settings)

        self.local_index_source = local_index_connector.LocalIndexSearchSource(
            source_operation=prep_operation
        )
        self.prep_operation = prep_operation

    def prepare(
        self, record: colrev.record.record_prep.PrepRecord
    ) -> colrev.record.record.Record:
        """Prepare the record metadata based on local-index"""

        # don't move to  jour_iss_number_year prep
        # because toc-retrieval relies on adequate toc items!
        if (
            Fields.VOLUME in record.data
            and Fields.NUMBER in record.data
            and not record.masterdata_is_curated()
        ):
            # Note : cannot use local_index as an attribute of PrepProcess
            # because it creates problems with multiprocessing
            fields_to_remove = self.local_index_source.local_index.get_fields_to_remove(
                record.get_data()
            )
            for field_to_remove in fields_to_remove:
                if field_to_remove in record.data:
                    record.remove_field(
                        key=field_to_remove,
                        not_missing_note=True,
                        source="local_index",
                    )

        self.local_index_source.prep_link_md(
            prep_operation=self.prep_operation, record=record
        )

        return record



#! /usr/bin/env python
"""Checker for container-title-abbreviated."""
from __future__ import annotations

import colrev.record.qm.quality_model
from colrev.constants import DefectCodes
from colrev.constants import Fields

# pylint: disable=too-few-public-methods


class ContainerTitleAbbreviatedChecker:
    """The ContainerTitleAbbreviatedChecker"""

    fields_to_check = [Fields.JOURNAL, Fields.BOOKTITLE]
    msg = DefectCodes.CONTAINER_TITLE_ABBREVIATED

    def __init__(
        self, quality_model: colrev.record.qm.quality_model.QualityModel
    ) -> None:
        self.quality_model = quality_model

    def run(self, *, record: colrev.record.record.Record) -> None:
        """Run the container-title-abbreviated checks"""

        for key in self.fields_to_check:
            if key not in record.data or record.ignored_defect(
                key=key, defect=self.msg
            ):
                continue
            if record.masterdata_is_curated():  # pragma: no cover
                continue

            if self.__container_title_abbreviated(record=record, key=key):
                record.add_field_provenance_note(key=key, note=self.msg)
            else:
                record.remove_field_provenance_note(key=key, note=self.msg)

    def __container_title_abbreviated(
        self, *, record: colrev.record.record.Record, key: str
    ) -> bool:
        if len(record.data[key]) < 6 and record.data[key].isupper():
            return True
        if key == Fields.BOOKTITLE and "Proc." in record.data[Fields.BOOKTITLE]:
            return True
        return False


def register(quality_model: colrev.record.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(ContainerTitleAbbreviatedChecker(quality_model))



#! /usr/bin/env python
"""SearchSource: OpenAlex"""
from __future__ import annotations

import typing
from multiprocessing import Lock
from pathlib import Path

import requests
import zope.interface
from pydantic import Field

import colrev.exceptions as colrev_exceptions
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
import colrev.record.record_prep
from colrev.constants import Fields
from colrev.constants import SearchSourceHeuristicStatus
from colrev.constants import SearchType
from colrev.packages.open_alex.src import open_alex_api

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.SearchSourceInterface)
class OpenAlexSearchSource:
    """OpenAlex API"""

    settings_class = colrev.package_manager.package_settings.DefaultSourceSettings
    endpoint = "colrev.open_alex"
    source_identifier = "openalex_id"
    search_types = [SearchType.MD]

    ci_supported: bool = Field(default=True)
    heuristic_status = SearchSourceHeuristicStatus.oni

    _open_alex_md_filename = Path("data/search/md_open_alex.bib")

    def __init__(
        self,
        *,
        source_operation: colrev.process.operation.Operation,
        settings: typing.Optional[dict] = None,
    ) -> None:
        self.review_manager = source_operation.review_manager
        # Note: not yet implemented
        # Note : once this is implemented, add "colrev.open_alex" to the default settings
        # if settings:
        #     # OpenAlex as a search_source
        #     self.search_source = self.settings_class(**settings)
        # else:
        # OpenAlex as an md-prep source
        open_alex_md_source_l = [
            s
            for s in self.review_manager.settings.sources
            if s.filename == self._open_alex_md_filename
        ]
        if open_alex_md_source_l:
            self.search_source = open_alex_md_source_l[0]
        else:
            self.search_source = colrev.settings.SearchSource(
                endpoint="colrev.open_alex",
                filename=self._open_alex_md_filename,
                search_type=SearchType.MD,
                search_parameters={},
                comment="",
            )

        self.open_alex_lock = Lock()

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for OpenAlex"""

        result = {"confidence": 0.0}

        return result

    @classmethod
    def add_endpoint(
        cls,
        operation: colrev.ops.search.Search,
        params: str,
    ) -> colrev.settings.SearchSource:
        """Add SearchSource as an endpoint (based on query provided to colrev search --add )"""

        raise colrev_exceptions.PackageParameterError(
            f"Cannot add OpenAlex endpoint with query {params}"
        )

    def check_availability(
        self, *, source_operation: colrev.process.operation.Operation
    ) -> None:
        """Check status (availability) of the OpenAlex API"""

    def _get_masterdata_record(
        self, *, record: colrev.record.record.Record
    ) -> colrev.record.record.Record:
        try:

            _, email = self.review_manager.get_committer()
            api = open_alex_api.OpenAlexAPI(email=email)
            retrieved_record = api.get_record(
                open_alex_id=record.data["colrev.open_alex.id"]
            )

            self.open_alex_lock.acquire(timeout=120)

            # Note : need to reload file because the object is not shared between processes
            open_alex_feed = self.search_source.get_api_feed(
                review_manager=self.review_manager,
                source_identifier=self.source_identifier,
                update_only=False,
                prep_mode=True,
            )

            open_alex_feed.add_update_record(retrieved_record)
            record.change_entrytype(
                new_entrytype=retrieved_record.data[Fields.ENTRYTYPE],
                qm=self.review_manager.get_qm(),
            )

            record.merge(
                retrieved_record,
                default_source=retrieved_record.data[Fields.ORIGIN][0],
            )
            open_alex_feed.save()
        except (
            colrev_exceptions.RecordNotParsableException,
            requests.exceptions.RequestException,
        ):
            pass
        except Exception as exc:
            raise exc
        finally:
            try:
                self.open_alex_lock.release()
            except ValueError:
                pass

        return record

    def prep_link_md(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
        save_feed: bool = True,
        timeout: int = 30,
    ) -> colrev.record.record.Record:
        """Retrieve masterdata from OpenAlex based on similarity with the record provided"""

        if "colrev.open_alex.id" in record.data:
            # Note: not yet implemented
            # https://github.com/OpenAPC/openapc-de/blob/master/python/import_dois.py
            # if len(record.data.get(Fields.TITLE, "")) < 35 and Fields.DOI not in record.data:
            #     return record
            # record = self._check_doi_masterdata(record=record)
            record = self._get_masterdata_record(record=record)

        return record

    def search(self, rerun: bool) -> None:
        """Run a search of OpenAlex"""

        # https://docs.openalex.org/api-entities/works

        raise NotImplementedError

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".bib":
            records = colrev.loader.load_utils.load(
                filename=self.search_source.filename,
                logger=self.review_manager.logger,
            )
            return records

        raise NotImplementedError

    def prepare(
        self, record: colrev.record.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.record.Record:
        """Source-specific preparation for OpenAlex"""

        return record



#!/usr/bin/env python3
"""Collection of utility functions"""
import operator
import pkgutil
import re
import typing
import unicodedata
from enum import Enum
from functools import reduce
from pathlib import Path
from pathlib import PosixPath

from jinja2 import Environment
from jinja2 import FunctionLoader
from jinja2.environment import Template

import colrev.exceptions as colrev_exceptions


def retrieve_package_file(*, template_file: Path, target: Path) -> None:
    """Retrieve a file from the CoLRev package"""
    try:
        filedata = pkgutil.get_data("colrev", str(template_file))
        if filedata:
            target.parent.mkdir(exist_ok=True, parents=True)
            with open(target, "w", encoding="utf8") as file:
                file.write(filedata.decode("utf-8"))
            return
    except FileNotFoundError:
        pass
    raise colrev_exceptions.TemplateNotAvailableError(str(template_file))


def get_package_file_content(
    *, module: str, filename: Path
) -> typing.Union[bytes, None]:
    """Get the content of a file in the CoLRev package"""
    return pkgutil.get_data(module, str(filename))


def inplace_change(*, filename: Path, old_string: str, new_string: str) -> None:
    """Replace a string in a file"""
    with open(filename, encoding="utf8") as file:
        content = file.read()
        if old_string not in content:
            return
    with open(filename, "w", encoding="utf8") as file:
        content = content.replace(old_string, new_string)
        file.write(content)


def get_template(template_path: str) -> Template:
    """Load a jinja template"""
    environment = Environment(
        loader=FunctionLoader(_load_jinja_template), autoescape=True
    )
    template = environment.get_template(template_path)
    return template


def _load_jinja_template(template_path: str) -> str:
    try:
        filedata_b = pkgutil.get_data("colrev", template_path)
        if filedata_b:
            filedata = filedata_b.decode("utf-8")
            filedata = filedata.replace("\n", "")
            filedata = filedata.replace("\r", "")
            filedata = filedata.replace("<br>", "\n")
            return filedata
    except FileNotFoundError:
        pass
    raise colrev_exceptions.TemplateNotAvailableError(template_path)


def remove_accents(input_str: str) -> str:
    """Replace the accents in a string"""

    nfkd_form = unicodedata.normalize("NFKD", input_str)
    wo_ac_list = [c for c in nfkd_form if not unicodedata.combining(c)]
    return "".join(wo_ac_list)


def percent_upper_chars(input_string: str) -> float:
    """Get the percentage of upper-case characters in a string"""

    input_string = re.sub(r"[^a-zA-Z]", "", input_string)
    if len(input_string) == 0:
        return 0.0
    return sum(map(str.isupper, input_string)) / len(input_string)


def custom_asdict_factory(data) -> dict:  # type: ignore
    """Custom asdict factory for (dataclass)object-to-dict conversion"""

    def convert_value(obj: object) -> object:
        if isinstance(obj, list):
            return [convert_value(el) for el in obj]
        if isinstance(obj, dict):
            return {k: convert_value(v) for k, v in obj.items()}
        if isinstance(obj, Enum):
            return obj.value
        if isinstance(obj, (Path, PosixPath)):
            return str(obj)
        if isinstance(obj, float):
            # Save 1.0 as 1 per default to avoid parsing issues
            # e.g., with the web ui
            if str(obj) == "1.0":  # pragma: no cover
                return 1
        return obj

    return {k: convert_value(v) for k, v in data.items()}


def load_complementary_material_keywords() -> list:
    """Load the list of keywords identifying complementary materials"""
    complementary_material_keywords = []
    filedata = get_package_file_content(
        module="colrev.env", filename=Path("complementary_material_keywords.txt")
    )
    if filedata:
        complementary_material_keywords = list(filedata.decode("utf-8").splitlines())

    return complementary_material_keywords


def load_complementary_material_strings() -> list:
    """Load the list of exact strings identifying complementary materials"""

    complementary_material_keywords = []
    filedata = get_package_file_content(
        module="colrev.env", filename=Path("complementary_material_strings.txt")
    )
    if filedata:
        complementary_material_keywords = list(filedata.decode("utf-8").splitlines())

    return complementary_material_keywords


def load_complementary_material_prefixes() -> list:
    """Load the list of prefixes identifying complementary materials"""

    complementary_material_keywords = []
    filedata = get_package_file_content(
        module="colrev.env", filename=Path("complementary_material_prefixes.txt")
    )
    if filedata:
        complementary_material_keywords = list(filedata.decode("utf-8").splitlines())

    return complementary_material_keywords


def get_by_path(root: dict, items: typing.List[str]) -> typing.Any:
    """Access a nested object in root by item sequence."""

    return reduce(operator.getitem, items, root)


def dict_set_nested(root: dict, keys: typing.List[str], value: typing.Any) -> None:
    """Set dict value by nested key, this works on empty dict"""
    for key in keys[:-1]:
        root = root.setdefault(key, {})
    root[keys[-1]] = value



#! /usr/bin/env python
"""CoLRev dedupe operation: identify and merge duplicate records."""
from __future__ import annotations

import typing

import pandas as pd

import colrev.ops.dedupe
from colrev.constants import Fields


def load_dedupe_false_positives(*, dedupe_operation: colrev.ops.dedupe.Dedupe) -> list:
    """Load the dedupe false positives marked in the Excel or txt file"""
    false_positives = []
    if dedupe_operation.dupe_file.is_file():
        dupes = pd.read_excel(dedupe_operation.dupe_file)
        dupes.fillna("", inplace=True)
        c_to_correct = dupes.loc[dupes["error"] != "", "cluster_id"].to_list()
        dupes = dupes[dupes["cluster_id"].isin(c_to_correct)]
        false_positives = (
            dupes.groupby(["cluster_id"], group_keys=False)[Fields.ID]
            .apply(list)
            .tolist()
        )
    return false_positives


def load_dedupe_false_negatives(*, dedupe_operation: colrev.ops.dedupe.Dedupe) -> list:
    """Load the dedupe false negatives marked in the Excel file"""

    false_negatives: typing.List[dict] = []
    if (
        dedupe_operation.non_dupe_file_xlsx.is_file()
        or dedupe_operation.non_dupe_file_txt.is_file()
    ):
        ids_to_merge = []
        if dedupe_operation.non_dupe_file_xlsx.is_file():
            non_dupes = pd.read_excel(dedupe_operation.non_dupe_file_xlsx)
            non_dupes.fillna("", inplace=True)
            c_to_correct = non_dupes.loc[
                non_dupes["error"] != "", "cluster_id"
            ].to_list()
            non_dupes = non_dupes[non_dupes["cluster_id"].isin(c_to_correct)]
            ids_to_merge = (
                non_dupes.groupby(["cluster_id"], group_keys=False)[Fields.ID]
                .apply(list)
                .tolist()
            )
        if dedupe_operation.non_dupe_file_txt.is_file():
            content = dedupe_operation.non_dupe_file_txt.read_text()
            ids_to_merge = [x.split(",") for x in content.splitlines()]
            for id_1, id_2 in ids_to_merge:
                print(f"{id_1} - {id_2}")

        false_negatives = []
        for id_list in ids_to_merge:
            if 2 == len(id_list):
                false_negatives.append(
                    {
                        "ID1": id_list[0],
                        "ID2": id_list[1],
                        "decision": "duplicate",
                    }
                )
            else:
                for i, idc in enumerate(id_list):
                    if 0 == i:
                        continue
                    false_negatives.append(
                        {
                            "ID1": id_list[0],
                            "ID2": idc,
                            "decision": "duplicate",
                        }
                    )
    return false_negatives



#! /usr/bin/env python
"""OCR as a PDF preparation operation"""
from __future__ import annotations

import os
import shutil
from pathlib import Path

import docker
import zope.interface
from pydantic import Field

import colrev.env.docker_manager
import colrev.env.utils
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Fields
from colrev.constants import PDFDefectCodes

# pylint: disable=too-few-public-methods
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.PDFPrepInterface)
class OCRMyPDF:
    """Prepare PDFs by applying OCR based on OCRmyPDF"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings
    ci_supported: bool = Field(default=False)
    OCRMYPDF_IMAGE = "jbarlow83/ocrmypdf:latest"

    def __init__(
        self,
        *,
        pdf_prep_operation: colrev.ops.pdf_prep.PDFPrep,
        settings: dict,
    ) -> None:
        self.settings = self.settings_class(**settings)
        self.review_manager = pdf_prep_operation.review_manager

        if not self.review_manager.in_ci_environment():
            colrev.env.docker_manager.DockerManager.build_docker_image(
                imagename=self.OCRMYPDF_IMAGE
            )
        pdf_prep_operation.docker_images_to_stop.append(self.OCRMYPDF_IMAGE)

    def _apply_ocr(
        self,
        *,
        record: colrev.record.record_pdf.PDFRecord,
    ) -> colrev.record.record_pdf.PDFRecord:
        pdf_path = self.review_manager.path / Path(record.data[Fields.FILE])
        non_ocred_filename = Path(str(pdf_path).replace(".pdf", "_no_ocr.pdf"))
        shutil.move(str(pdf_path), str(non_ocred_filename))
        orig_path = (
            pdf_path.parents[0] if pdf_path.is_file() else self.review_manager.paths.pdf
        )

        # options = ""
        # if rotate:
        #     options = options + '--rotate-pages '
        # if deskew:
        #     options = options + '--deskew '
        docker_home_path = Path("/home/docker")

        non_ocr_fname = docker_home_path / non_ocred_filename.name
        docker_output_fname = docker_home_path / pdf_path.name
        args = (
            f"--force-ocr --jobs 4 -l eng {non_ocr_fname.as_posix()} "
            f"{docker_output_fname.as_posix()}"
        )

        client = docker.from_env(timeout=120)
        client.containers.run(
            image=self.OCRMYPDF_IMAGE,
            command=args,
            auto_remove=True,
            user=f"{os.geteuid()}:{os.getegid()}",
            volumes=[f"{orig_path}:/home/docker"],
        )

        record.add_field_provenance_note(
            key=Fields.FILE, note="pdf_processed with OCRMYPDF"
        )
        record.set_text_from_pdf()
        return record

    def prep_pdf(
        self,
        record: colrev.record.record_pdf.PDFRecord,
        pad: int,  # pylint: disable=unused-argument
    ) -> dict:
        """Prepare the PDF by applying OCR"""

        if (
            Fields.FILE not in record.data
            or not record.data[Fields.FILE].endswith(".pdf")
            or PDFDefectCodes.NO_TEXT_IN_PDF not in record.defects("file")
        ):
            return record.data

        self.review_manager.report_logger.info(f"apply_ocr({record.data[Fields.ID]})")
        record = self._apply_ocr(record=record)

        return record.data



#!/usr/bin/env python3
"""Hook for reporting in CoLRev projects"""
import sys
from pathlib import Path

import colrev.review_manager
from colrev.constants import ExitCodes


def main() -> int:
    """Main entrypoint for the reporting"""

    print(sys.argv)
    msgfile = Path(sys.argv[1])

    review_manager = colrev.review_manager.ReviewManager()
    review_manager.report(msg_file=msgfile)

    return ExitCodes.SUCCESS


if __name__ == "__main__":
    raise SystemExit(main())



#! /usr/bin/env python
"""Crossref API"""
import contextlib
import re
import typing
import urllib
from datetime import timedelta
from importlib.metadata import version
from pathlib import Path
from time import sleep

import requests
import requests_cache
from rapidfuzz import fuzz

import colrev.env.environment_manager
import colrev.exceptions as colrev_exceptions
import colrev.record.record_prep
from colrev.constants import Colors
from colrev.constants import Fields
from colrev.constants import Filepaths
from colrev.packages.crossref.src import record_transformer

LIMIT = 1000
MAXOFFSET = 10000

SESSION = requests_cache.CachedSession(
    str(Filepaths.LOCAL_ENVIRONMENT_DIR / Path("crossref_cache.sqlite")),
    backend="sqlite",
    expire_after=timedelta(days=30),
)


class CrossrefAPIError(Exception):
    """Crossref API Error"""


class MaxOffsetError(CrossrefAPIError):
    """Max Offset Error"""


# pylint: disable=too-few-public-methods
class HTTPRequest:
    """HTTP Request"""

    def __init__(self, *, timeout: int) -> None:
        self.rate_limits = {"x-rate-limit-limit": 50, "x-rate-limit-interval": 1}
        self.timeout = timeout

    def _update_rate_limits(self, headers: dict) -> None:

        with contextlib.suppress(ValueError):
            self.rate_limits["x-rate-limit-limit"] = int(
                headers.get("x-rate-limit-limit", 50)
            )

        with contextlib.suppress(ValueError):
            interval_value = int(headers.get("x-rate-limit-interval", "1s")[:-1])

        interval_scope = headers.get("x-rate-limit-interval", "1s")[-1]

        if interval_scope == "m":
            interval_value = interval_value * 60

        if interval_scope == "h":
            interval_value = interval_value * 60 * 60

        self.rate_limits["x-rate-limit-interval"] = interval_value

    def _get_throttling_time(self) -> float:
        """"""
        return (
            self.rate_limits["x-rate-limit-interval"]
            / self.rate_limits["x-rate-limit-limit"]
        )

    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-positional-arguments
    def retrieve(
        self,
        endpoint: str,
        headers: dict,
        data: typing.Optional[dict] = None,
        only_headers: bool = False,
        skip_throttle: bool = False,
        cache: bool = True,
    ) -> requests.Response:
        """Retrieve data from a given endpoint."""

        if only_headers is True:
            return requests.head(endpoint, timeout=2)

        if cache:
            result = SESSION.get(
                endpoint, params=data, timeout=self.timeout, headers=headers
            )
        else:
            result = requests.get(
                endpoint, params=data, timeout=self.timeout, headers=headers
            )

        if not skip_throttle:
            self._update_rate_limits(result.headers)
            sleep(self._get_throttling_time())

        return result


class Endpoint:
    """Endpoint"""

    CURSOR_AS_ITER_METHOD = False

    def __init__(
        self,
        request_url: str,
        *,
        email: str = "",
        crossref_plus_token: str = "",
    ) -> None:

        self.retrieve = HTTPRequest(timeout=60).retrieve

        self.headers = {
            "user-agent": f"colrev/{version('colrev')} "
            + f"(https://github.com/CoLRev-Environment/colrev; mailto:{email})"
        }
        self.crossref_plus_token = crossref_plus_token
        if crossref_plus_token:
            self.headers["Crossref-Plus-API-Token"] = self.crossref_plus_token
        self.request_url = request_url
        self.request_params: typing.Dict[str, str] = {}
        self.timeout = 60

    @property
    def _rate_limits(self) -> dict:
        request_url = str(self.request_url)

        result = self.retrieve(
            request_url,
            only_headers=True,
            headers=self.headers,
            skip_throttle=True,
        )

        return {
            "x-rate-limit-limit": result.headers.get("x-rate-limit-limit", "undefined"),
            "x-rate-limit-interval": result.headers.get(
                "x-rate-limit-interval", "undefined"
            ),
        }

    def _escaped_pagging(self) -> dict:
        escape_pagging = ["offset", "rows"]
        request_params = dict(self.request_params)

        for item in escape_pagging:
            with contextlib.suppress(KeyError):
                del request_params[item]

        return request_params

    @property
    def version(self) -> str:
        """API version."""

        request_params = dict(self.request_params)
        request_url = str(self.request_url)

        result = self.retrieve(
            request_url,
            data=request_params,
            headers=self.headers,
        ).json()

        return result["message-version"]

    def get_nr(self) -> int:
        """Retrieve the total number of records resulting from a query."""
        request_params = dict(self.request_params)
        request_url = str(self.request_url)
        request_params["rows"] = "0"

        try:
            result = self.retrieve(
                request_url,
                data=request_params,
                headers=self.headers,
            ).json()
        except requests.exceptions.RequestException as exc:
            raise colrev_exceptions.ServiceNotAvailableException(
                f"Crossref ({Colors.ORANGE}check https://status.crossref.org/{Colors.END})"
            ) from exc

        return int(result["message"]["total-results"])

    @property
    def url(self) -> str:
        """Retrieve the url that will be used as a HTTP request."""

        request_params = self._escaped_pagging()
        sorted_request_params = sorted(request_params.items())

        req = requests.Request(
            "get", self.request_url, params=sorted_request_params
        ).prepare()

        return req.url

    # pylint: disable=too-many-branches
    # pylint: disable=too-many-return-statements
    def __iter__(self) -> typing.Iterator[dict]:

        request_url = str(self.request_url)

        if request_url.startswith("https://api.crossref.org/works/10."):
            result = self.retrieve(
                request_url,
                headers=self.headers,
            )

            if result.status_code == 404:
                return

            result = result.json()

            yield result["message"]

            return

        if self.CURSOR_AS_ITER_METHOD is True:
            request_params = dict(self.request_params)
            request_params["cursor"] = "*"
            request_params["rows"] = str(LIMIT)
            while True:

                result = self.retrieve(
                    request_url,
                    data=request_params,
                    headers=self.headers,
                    cache=False,
                )

                if result.status_code == 404:
                    return

                result = result.json()

                if len(result["message"]["items"]) == 0:
                    return

                yield from result["message"]["items"]

                request_params["cursor"] = result["message"]["next-cursor"]

        else:
            request_params = dict(self.request_params)
            request_params["offset"] = "0"
            request_params["rows"] = str(LIMIT)
            while True:
                result = self.retrieve(
                    request_url,
                    data=request_params,
                    headers=self.headers,
                )

                if result.status_code == 404:
                    return

                result = result.json()

                if len(result["message"]["items"]) == 0:
                    return

                yield from result["message"]["items"]

                request_params["offset"] = str(int(request_params["offset"]) + LIMIT)

                if int(request_params["offset"]) >= MAXOFFSET:
                    msg = "Offset exceded the max offset of %d"
                    raise MaxOffsetError(msg, MAXOFFSET)


class CrossrefAPI:
    """Crossref API"""

    ISSN_REGEX = r"^\d{4}-?\d{3}[\dxX]$"
    YEAR_SCOPE_REGEX = r"^\d{4}-\d{4}$"

    # https://github.com/CrossRef/rest-api-doc
    _api_url = "https://api.crossref.org/"

    last_updated: str = ""

    _availability_exception_message = (
        f"Crossref ({Colors.ORANGE}check https://status.crossref.org/{Colors.END})"
    )

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        *,
        params: dict,
        rerun: bool = False,
    ):
        self.params = params

        _, self.email = (
            colrev.env.environment_manager.EnvironmentManager.get_name_mail_from_git()
        )
        self.rerun = rerun

    def check_availability(self, raise_service_not_available: bool = True) -> None:
        """Check the availability of the API"""

        try:
            # pylint: disable=duplicate-code
            test_rec = {
                Fields.DOI: "10.17705/1cais.04607",
                Fields.AUTHOR: "Schryen, Guido and Wagner, Gerit and Benlian, Alexander "
                "and Paré, Guy",
                Fields.TITLE: "A Knowledge Development Perspective on Literature Reviews: "
                "Validation of a new Typology in the IS Field",
                Fields.ID: "SchryenEtAl2021",
                Fields.JOURNAL: "Communications of the Association for Information Systems",
                Fields.ENTRYTYPE: "article",
            }

            returned_record = self.crossref_query(
                record_input=colrev.record.record_prep.PrepRecord(test_rec),
                jour_vol_iss_list=False,
            )[0]

            if 0 != len(returned_record.data):
                assert returned_record.data[Fields.TITLE] == test_rec[Fields.TITLE]
                assert returned_record.data[Fields.AUTHOR] == test_rec[Fields.AUTHOR]
            else:
                if raise_service_not_available:
                    raise colrev_exceptions.ServiceNotAvailableException(
                        self._availability_exception_message
                    )
        except (requests.exceptions.RequestException, IndexError) as exc:
            print(exc)
            if raise_service_not_available:
                raise colrev_exceptions.ServiceNotAvailableException(
                    self._availability_exception_message
                ) from exc

    def get_url(self) -> str:
        """Get the url for the Crossref API"""

        if "url" not in self.params:
            raise ValueError("No url in params")

        url = self.params["url"]
        if not self.rerun and self.last_updated:
            # see https://api.staging.crossref.org/swagger-ui/
            # index.html#/Journals/get_journals__issn__works
            # "Notes on incremental metadata updates"

            last_updated = self.last_updated.split("T", maxsplit=1)[0]
            url = url + f"?filter=from-index-date:{last_updated}"

        return url

    def get_len_total(self) -> int:
        """Get the total number of records from Crossref based on the parameters"""

        endpoint = Endpoint(self.params["url"], email=self.email)
        return endpoint.get_nr()

    def get_len(self) -> int:
        """Get the number of records from Crossref based on the parameters"""

        endpoint = Endpoint(self.get_url(), email=self.email)
        return endpoint.get_nr()

    def get_records(self) -> typing.Iterator[colrev.record.record.Record]:
        """Get records from Crossref based on the parameters"""

        url = self.get_url()

        endpoint = Endpoint(url, email=self.email)
        if self.get_len() > 10000:
            endpoint.CURSOR_AS_ITER_METHOD = True

        try:
            for item in endpoint:
                try:
                    yield record_transformer.json_to_record(item=item)
                except colrev_exceptions.RecordNotParsableException:
                    continue
        except requests.exceptions.RequestException as exc:
            raise colrev_exceptions.ServiceNotAvailableException(
                self._availability_exception_message
            ) from exc

    def query_doi(self, *, doi: str) -> colrev.record.record_prep.PrepRecord:
        """Get records from Crossref based on a doi query"""

        try:
            endpoint = Endpoint(self._api_url + "works/" + doi, email=self.email)

            crossref_query_return = next(iter(endpoint))

            if crossref_query_return is None:
                raise colrev_exceptions.RecordNotFoundInPrepSourceException(
                    msg="Record not found in crossref (based on doi)"
                )

            retrieved_record = record_transformer.json_to_record(
                item=crossref_query_return
            )
            return retrieved_record

        except (requests.exceptions.RequestException, StopIteration) as exc:
            raise colrev_exceptions.RecordNotFoundInPrepSourceException(
                msg="Record not found in crossref (based on doi)"
            ) from exc

    def _get_similarity(
        self, *, record: colrev.record.record.Record, retrieved_record_dict: dict
    ) -> float:
        title_similarity = fuzz.partial_ratio(
            retrieved_record_dict.get(Fields.TITLE, "NA").lower(),
            record.data.get(Fields.TITLE, "").lower(),
        )
        container_similarity = fuzz.partial_ratio(
            colrev.record.record_prep.PrepRecord(retrieved_record_dict)
            .get_container_title()
            .lower(),
            record.get_container_title().lower(),
        )
        weights = [0.6, 0.4]
        similarities = [title_similarity, container_similarity]

        similarity = sum(similarities[g] * weights[g] for g in range(len(similarities)))
        return similarity

    def _create_query_url(
        self, *, record: colrev.record.record.Record, jour_vol_iss_list: bool
    ) -> str:
        if jour_vol_iss_list:
            if not all(
                x in record.data for x in [Fields.JOURNAL, Fields.VOLUME, Fields.NUMBER]
            ):
                raise colrev_exceptions.NotEnoughDataToIdentifyException
            params = {}
            container_title = re.sub(r"[\W]+", " ", record.data[Fields.JOURNAL])
            params["query.container-title"] = container_title.replace("_", " ")

            query_field = ""
            if Fields.VOLUME in record.data:
                query_field = record.data[Fields.VOLUME]
            if Fields.NUMBER in record.data:
                query_field = query_field + "+" + record.data[Fields.NUMBER]
            params["query"] = query_field
            # params["rows"]  = "50"

        else:
            if Fields.TITLE not in record.data:
                raise colrev_exceptions.NotEnoughDataToIdentifyException()

            params = {}
            if not isinstance(record.data.get(Fields.YEAR, ""), str) or not isinstance(
                record.data.get(Fields.TITLE, ""), str
            ):
                print("year or title field not a string")
                print(record.data)
                raise AssertionError

            bibl = (
                record.data[Fields.TITLE].replace("-", "_")
                + " "
                + record.data.get(Fields.YEAR, "")
            )
            bibl = re.sub(r"[\W]+", "", bibl.replace(" ", "_"))
            params["query.bibliographic"] = bibl.replace("_", " ").rstrip("+")

            container_title = record.get_container_title()
            if "." not in container_title:
                container_title = container_title.replace(" ", "_")
                container_title = re.sub(r"[\W]+", "", container_title)
                params["query.container-title"] = container_title.replace("_", " ")

            author_last_names = [
                x.split(",")[0]
                for x in record.data.get(Fields.AUTHOR, "").split(" and ")
            ]
            author_string = " ".join(author_last_names)
            author_string = re.sub(r"[\W]+", "", author_string.replace(" ", "_"))
            params["query.author"] = author_string.replace("_", " ")
            # params["rows"] = "15"

        url = self._api_url + "works?" + urllib.parse.urlencode(params)
        return url

    def crossref_query(
        self,
        *,
        record_input: colrev.record.record.Record,
        jour_vol_iss_list: bool = False,
    ) -> list:
        """Retrieve records from Crossref based on a query"""

        record = record_input.copy_prep_rec()
        record_list, most_similar, most_similar_record = [], 0.0, {}

        url = self._create_query_url(record=record, jour_vol_iss_list=jour_vol_iss_list)

        endpoint = Endpoint(url, email=self.email)
        if jour_vol_iss_list:
            endpoint.request_params["rows"] = "50"
        else:
            endpoint.request_params["rows"] = "15"

        counter = 0
        while True:
            try:
                item = next(iter(endpoint), None)
            except requests.exceptions.RequestException as exc:
                raise colrev_exceptions.ServiceNotAvailableException(
                    f"Crossref ({Colors.ORANGE}check https://status.crossref.org/{Colors.END})"
                ) from exc
            if item is None:
                break
            try:
                retrieved_record = record_transformer.json_to_record(item=item)
                similarity = self._get_similarity(
                    record=record, retrieved_record_dict=retrieved_record.data
                )

                if jour_vol_iss_list:
                    record_list.append(retrieved_record)
                elif most_similar < similarity:
                    most_similar = similarity
                    most_similar_record = retrieved_record.get_data()
            except colrev_exceptions.RecordNotParsableException:
                pass
            counter += 1
            if jour_vol_iss_list and counter > 200:
                break
            if not jour_vol_iss_list and counter > 5:
                break

        if not jour_vol_iss_list:
            if most_similar_record:
                record_list = [
                    colrev.record.record_prep.PrepRecord(most_similar_record)
                ]

        return record_list



#! /usr/bin/env python
"""Colrev curated data as part of the data operations"""
from __future__ import annotations

import collections
import os
import typing
from pathlib import Path

import pandas as pd
import zope.interface
from pydantic import BaseModel
from pydantic import Field

import colrev.env.utils
import colrev.exceptions as colrev_exceptions
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Fields
from colrev.constants import RecordState


class ColrevCurationSettings(
    colrev.package_manager.package_settings.DefaultSettings, BaseModel
):
    """Colrev Curation settings"""

    endpoint: str
    version: str
    curation_url: str
    curated_masterdata: bool
    masterdata_restrictions: dict
    curated_fields: list


@zope.interface.implementer(colrev.package_manager.interfaces.DataInterface)
class ColrevCuration:
    """CoLRev Curation"""

    settings: ColrevCurationSettings
    ci_supported: bool = Field(default=True)

    settings_class = ColrevCurationSettings

    def __init__(
        self,
        *,
        data_operation: colrev.ops.data.Data,
        settings: dict,
    ) -> None:
        # Set default values (if necessary)
        if "version" not in settings:
            settings["version"] = "0.1"

        self.settings = ColrevCurationSettings(**settings)
        self.review_manager = data_operation.review_manager

        self.data_operation = data_operation

    # pylint: disable=unused-argument
    @classmethod
    def add_endpoint(cls, operation: colrev.ops.data.Data, params: str) -> None:
        """Add as an endpoint"""

        add_source = {
            "endpoint": "colrev.colrev_curation",
            "version": "0.1",
            "curation_url": "TODO",
            "curated_masterdata": True,
            "masterdata_restrictions": {
                1900: {
                    Fields.ENTRYTYPE: "article",
                    Fields.JOURNAL: "TODO",
                    Fields.VOLUME: True,
                    Fields.NUMBER: True,
                }
            },
            "curated_fields": [Fields.DOI, Fields.URL],
        }

        operation.review_manager.settings.data.data_package_endpoints.append(add_source)
        operation.review_manager.save_settings()
        operation.review_manager.dataset.create_commit(
            msg=f"Add {operation.type} colrev.curation",
        )

    def _get_stats(
        self,
        *,
        records: dict,
        sources: list,
    ) -> dict:
        # pylint: disable=too-many-branches

        stats: dict = {}
        for record_dict in records.values():
            r_status = str(record_dict[Fields.STATUS])
            if r_status == "rev_prescreen_excluded":
                continue
            if record_dict[Fields.STATUS] in RecordState.get_post_x_states(
                state=RecordState.md_processed
            ):
                r_status = str(RecordState.md_processed)
            elif record_dict[Fields.STATUS] in RecordState.get_post_x_states(
                state=RecordState.pdf_prepared
            ):
                r_status = str(RecordState.pdf_prepared)
            else:
                r_status = str(RecordState.md_imported)

            if Fields.JOURNAL in record_dict:
                key = (
                    f"{record_dict.get('year', '-')}-"
                    f"{record_dict.get('volume', '-')}-"
                    f"{record_dict.get('number', '-')}"
                )
            elif Fields.BOOKTITLE in record_dict:
                key = record_dict.get(Fields.YEAR, "-")
            else:
                self.review_manager.logger.error(f"TOC not supported: {record_dict}")
                continue
            if not all(
                source in [o.split("/")[0] for o in record_dict[Fields.ORIGIN]]
                for source in sources
            ):
                if key in stats:
                    if "all_merged" in stats[key]:
                        stats[key]["all_merged"] = "NO"
                    else:
                        stats[key]["all_merged"] = "NO"
                else:
                    stats[key] = {"all_merged": "NO"}

            for origin in record_dict[Fields.ORIGIN]:
                source = origin.split("/")[0]
                if key in stats:
                    if source in stats[key]:
                        if r_status in stats[key][source]:
                            stats[key][source][r_status] += 1
                        else:
                            stats[key][source][r_status] = 1
                    else:
                        stats[key][source] = {r_status: 1}
                else:
                    stats[key] = {source: {r_status: 1}}
        return stats

    def _get_stats_markdown_table(self, *, stats: dict, sources: list) -> str:
        cell_width = 16
        sources += ["all_merged"]
        output = "|TOC".ljust(cell_width - 1, " ") + "|"
        sub_header_lines = "|".ljust(cell_width - 1, "-") + "|"
        for source in sources:
            output += f"{source}".ljust(cell_width, " ") + "|"
            sub_header_lines += "".ljust(cell_width, "-") + "|"

        output += "\n" + sub_header_lines
        ordered_stats = collections.OrderedDict(sorted(stats.items(), reverse=True))

        output += self._get_stats_markdown_table_cell(
            ordered_stats=ordered_stats, sources=sources, cell_width=cell_width
        )

        output += "\n\nLegend: *md_imported*, md_processed, **pdf_prepared**"
        return output

    def _get_stats_markdown_table_cell(
        self, *, ordered_stats: collections.OrderedDict, sources: list, cell_width: int
    ) -> str:
        output = ""
        for key, row in ordered_stats.items():
            output += f"\n|{key}".ljust(cell_width, " ") + "|"
            for source in sources:
                if source != "all_merged":
                    if source in row:
                        cell_text = ""
                        if "md_imported" in row[source]:
                            cell_text += f"*{row[source]['md_imported']}*,"
                        else:
                            cell_text += "-,"
                        if "md_processed" in row[source]:
                            cell_text += f"{row[source]['md_processed']}"
                        else:
                            cell_text += "-"
                        if "pdf_prepared" in row[source]:
                            cell_text += f",**{row[source]['pdf_prepared']}**"
                        else:
                            cell_text += ",-"

                        output += cell_text.rjust(cell_width, " ") + "|"
                    else:
                        output += "-".rjust(cell_width, " ") + "|"
                else:
                    output += row.get("all_merged", "").rjust(cell_width, " ") + "|"
        return output

    def _update_table_in_readme(
        self,
        *,
        markdown_output: str,
    ) -> None:
        table_summary_tag = "<!-- TABLE_SUMMARY -->"
        readme_path = self.review_manager.paths.readme
        with open(readme_path, "r+b") as file:
            appended = False
            seekpos = file.tell()
            line = file.readline()
            while line:
                if table_summary_tag.encode("utf-8") in line:
                    line = file.readline()
                    while (
                        b"Legend: *md_imported*, md_processed" not in line and line
                    ):  # replace: drop the current record
                        line = file.readline()

                    remaining = file.read()
                    file.seek(seekpos)
                    file.write(table_summary_tag.encode("utf-8"))
                    file.write(b"\n\n")
                    file.write(markdown_output.encode("utf-8"))
                    file.write(b"\n")
                    seekpos = file.tell()
                    file.flush()
                    os.fsync(file)
                    file.write(remaining)
                    file.truncate()  # if the replacement is shorter...
                    file.seek(seekpos)
                    appended = True

                seekpos = file.tell()
                line = file.readline()
            if not appended:
                file.write(b"\n")
                file.write(table_summary_tag.encode("utf-8"))
                file.write(b"\n\n")
                file.write(markdown_output.encode("utf-8"))
                file.write(b"\n")

    def _update_stats_in_readme(
        self,
        *,
        records: dict,
        silent_mode: bool,
    ) -> None:
        if not silent_mode:
            self.review_manager.logger.info("Calculate statistics for readme")

        # alternatively: get sources from search_sources.filename (name/stem?)
        sources = []
        for record_dict in records.values():
            for origin in record_dict[Fields.ORIGIN]:
                source = origin.split("/")[0]
                if source not in sources:
                    sources.append(source)

        stats = self._get_stats(records=records, sources=sources)
        markdown_output = self._get_stats_markdown_table(stats=stats, sources=sources)
        self._update_table_in_readme(markdown_output=markdown_output)
        self.review_manager.dataset.add_changes(self.review_manager.paths.README_FILE)

    def _source_comparison(self, *, silent_mode: bool) -> None:
        """Exports a table to support analyses of records that are not
        in all sources (for curated repositories)"""
        dedupe_dir = self.review_manager.paths.dedupe
        dedupe_dir.mkdir(exist_ok=True, parents=True)
        source_comparison_xlsx = dedupe_dir / Path("source_comparison.xlsx")

        source_filenames = [
            str(x.filename) for x in self.review_manager.settings.sources
        ]
        if not silent_mode:
            print("sources: " + ",".join([str(x) for x in source_filenames]))

        records = self.review_manager.dataset.load_records_dict()
        records = {
            k: v
            for k, v in records.items()
            if not all(x in ";".join(v[Fields.ORIGIN]) for x in str(source_filenames))
        }
        if len(records) == 0:
            if not silent_mode:
                print("No records unmatched")
            return

        for record in records.values():
            origins = record[Fields.ORIGIN]
            for source_filename in source_filenames:
                if not any(source_filename in origin for origin in origins):
                    record[source_filename] = ""
                else:
                    record[source_filename] = [
                        origin for origin in origins if source_filename in origin
                    ][0]
            record["merge_with"] = ""

        records_df = pd.DataFrame.from_records(list(records.values()))
        records_df.to_excel(source_comparison_xlsx, index=False)
        if not silent_mode:
            print(f"Exported {source_comparison_xlsx}")

    def update_data(
        self,
        records: dict,
        synthesized_record_status_matrix: dict,  # pylint: disable=unused-argument
        silent_mode: bool,
    ) -> None:
        """Update the CoLRev curation"""

        if self.settings.curated_masterdata:
            self._update_stats_in_readme(
                records=records,
                silent_mode=silent_mode,
            )
            self._source_comparison(silent_mode=silent_mode)

    def update_record_status_matrix(
        self,
        synthesized_record_status_matrix: dict,
        endpoint_identifier: str,
    ) -> None:
        """Update the record_status_matrix"""

        for item in synthesized_record_status_matrix:
            synthesized_record_status_matrix[item][endpoint_identifier] = True

    def get_advice(
        self,
    ) -> dict:
        """Get advice on the next steps (for display in the colrev status)"""

        records = self.review_manager.dataset.load_records_dict()
        advice = {
            "msg": "TODO (add curation-specific advice...)",
            "detailed_msg": "TODO",
        }

        records_missing_languages = [
            r[Fields.ID] for r in records.values() if Fields.LANGUAGE not in r
        ]
        if records_missing_languages:
            advice = {
                "msg": "Curation: Add language field to all records",
                "detailed_msg": "records missing language field: "
                + f"({','.join(records_missing_languages)})",
            }

        identical_colrev_ids: typing.Dict[str, list] = {}
        non_identifiable_records = []
        for record_dict in records.values():
            try:
                if record_dict[Fields.STATUS] not in RecordState.get_post_x_states(
                    state=RecordState.md_prepared
                ):
                    continue
                if record_dict[Fields.STATUS] == RecordState.rev_prescreen_excluded:
                    continue
                cid = colrev.record.record.Record(record_dict).get_colrev_id(
                    assume_complete=True
                )
                if cid in identical_colrev_ids:
                    identical_colrev_ids[cid] = identical_colrev_ids[cid] + [
                        record_dict[Fields.ID]
                    ]
                else:
                    identical_colrev_ids[cid] = [record_dict[Fields.ID]]
            except colrev_exceptions.NotEnoughDataToIdentifyException:
                non_identifiable_records.append(record_dict[Fields.ID])

        identical_colrev_ids = {
            k: v for k, v in identical_colrev_ids.items() if len(v) > 1
        }
        if identical_colrev_ids:
            advice = {
                "msg": "Curation: resolve records with identical colrev_ids:\n      - "
                + "\n      - ".join(",".join(x) for x in identical_colrev_ids.values()),
                "detailed_msg": "records missing language field: ",
            }

        if not self.settings.masterdata_restrictions:
            advice = {
                "msg": "Curation: masterdata_restrictions not set. See "
                + "https://github.com/CoLRev-Environment/colrev/blob/main/"
                + "colrev/packages/data/colrev_curation.md for details.",
                "detailed_msg": "records missing language field: ",
            }

        return advice



"""Package for github_search."""

__author__ = "Philipp Kasimir "
__email__ = "philipp-william.kasimir@stud.uni-bamberg.de"

__author__ = "Kolja Rinne"
__email__ = "kolja.rinne@stud.uni-bamberg.de"

__author__ = "Karl Schickmann"
__email__ = "karl.schnickmann@stud.uni-bamberg.de"


__author__ = "Chris Vierath"
__email__ = "chris-norman.vierath@stud.uni-bamberg.de"



#! /usr/bin/env python
"""Service to detect languages and handle language codes"""
from __future__ import annotations

import re

import pycountry
from lingua import LanguageDetectorBuilder  # pylint: disable=no-name-in-module

import colrev.exceptions as colrev_exceptions
import colrev.record.record
from colrev.constants import Fields


class LanguageService:
    """Service to detect languages and handle language codes"""

    _eng_false_negatives = ["editorial", "introduction"]

    def __init__(self) -> None:
        # Note : Lingua is tested/evaluated relative to other libraries:
        # https://github.com/pemistahl/lingua-py
        # It performs particularly well for short strings (single words/word pairs)
        # The langdetect library is non-deterministic, especially for short strings
        # https://pypi.org/project/langdetect/

        self._lingua_language_detector = (
            LanguageDetectorBuilder.from_all_languages_with_latin_script().build()
        )

        # Language formats: ISO 639-1 standard language codes
        # https://pypi.org/project/langcodes/
        # https://github.com/flyingcircusio/pycountry

        self._lang_code_mapping = {}
        for country in pycountry.languages:
            self._lang_code_mapping[country.name.lower()] = country.alpha_3

    # pylint: disable=too-many-return-statements
    # pylint: disable=too-many-branches
    def _determine_alphabet(self, str_to_check: str) -> str:
        assert len(str_to_check) != 0

        str_to_check = re.sub(r"[\s\d\.\:]*", "", str_to_check)

        nr_greek_letters = 0
        nr_hangul_characters = 0
        nr_cyrillic_characters = 0
        nr_hebrew_characters = 0
        nr_arabic_characters = 0
        nr_chinese_characters = 0
        for character in str_to_check:
            if "\u0370" <= character <= "\u03FF" or "\u1F00" <= character <= "\u1FFF":
                nr_greek_letters += 1
            elif "\uAC00" <= character <= "\uD7A3":
                nr_hangul_characters += 1
            elif "\u0400" <= character <= "\u04FF" or "\u0500" <= character <= "\u052F":
                nr_cyrillic_characters += 1
            elif "\u0590" <= character <= "\u05FF" or "\uFB1D" <= character <= "\uFB4F":
                nr_hebrew_characters += 1
            elif (
                "\u0600" <= character <= "\u06FF"
                or "\u0750" <= character <= "\u077F"
                or "\u08A0" <= character <= "\u08FF"
                or "\uFB50" <= character <= "\uFDFF"
                or "\uFE70" <= character <= "\uFEFF"
            ):
                nr_arabic_characters += 1
            elif "\u4E00" <= character <= "\u9FFF" or "\u3400" <= character <= "\u4DBF":
                nr_chinese_characters += 1
        if nr_greek_letters / len(str_to_check) > 0.75:
            return "ell"
        if nr_hangul_characters / len(str_to_check) > 0.75:
            return "kor"
        if nr_cyrillic_characters / len(str_to_check) > 0.75:
            return "rus"
        if nr_hebrew_characters / len(str_to_check) > 0.75:
            return "heb"
        if nr_arabic_characters / len(str_to_check) > 0.75:
            return "ara"
        if nr_chinese_characters / len(str_to_check) > 0.75:
            return "chi"
        return ""  # pragma: no cover

    def compute_language(self, *, text: str) -> str:
        """Compute the most likely language code"""

        if text.lower() in self._eng_false_negatives:
            return "eng"

        language = self._lingua_language_detector.detect_language_of(text)

        if language:
            # There are too many errors/classifying papers as latin
            if language.iso_code_639_3.name.lower() == "lat":
                return ""
            return language.iso_code_639_3.name.lower()

        return self._determine_alphabet(text)

    def compute_language_confidence_values(self, *, text: str) -> list:
        """Computes the most likely languages of a string and their language codes"""

        if text.lower() in self._eng_false_negatives:
            return [("eng", 1.0)]

        predictions = self._lingua_language_detector.compute_language_confidence_values(
            text=text
        )
        predictions_unified = []
        for prediction in predictions:
            lang = prediction.language
            conf = prediction.value
            predictions_unified.append((lang.iso_code_639_3.name.lower(), conf))

        return predictions_unified

    def validate_iso_639_3_language_codes(self, *, lang_code_list: list) -> None:
        """Validates whether a list of language codes complies with the ISO 639-3 standard"""

        assert isinstance(lang_code_list, list)

        invalid_language_codes = [x for x in lang_code_list if 3 != len(x)]
        if invalid_language_codes:
            raise colrev_exceptions.InvalidLanguageCodeException(
                invalid_language_codes=invalid_language_codes
            )

    def unify_to_iso_639_3_language_codes(
        self, *, record: colrev.record.record.Record
    ) -> None:
        """Unifies a language_code string to the ISO 639-3 standard"""

        if Fields.LANGUAGE not in record.data:
            return

        if record.data[Fields.LANGUAGE].lower() in ["en"]:
            record.data[Fields.LANGUAGE] = "eng"

        elif record.data[Fields.LANGUAGE].lower() in ["fr"]:
            record.data[Fields.LANGUAGE] = "fra"

        elif record.data[Fields.LANGUAGE].lower() in ["ar"]:
            record.data[Fields.LANGUAGE] = "ara"

        elif record.data[Fields.LANGUAGE].lower() in ["de"]:
            record.data[Fields.LANGUAGE] = "deu"

        if len(record.data[Fields.LANGUAGE]) != 3:
            if record.data[Fields.LANGUAGE].lower() in self._lang_code_mapping:
                record.data[Fields.LANGUAGE] = self._lang_code_mapping[
                    record.data[Fields.LANGUAGE].lower()
                ]

        self.validate_iso_639_3_language_codes(
            lang_code_list=[record.data[Fields.LANGUAGE]]
        )



#! /usr/bin/env python
"""CoLRev pull operation: Pull project and records."""
from __future__ import annotations

import colrev.process.operation
from colrev.constants import Colors
from colrev.constants import OperationsType

# pylint: disable=too-few-public-methods

CHANGE_COUNTER = None


class Pull(colrev.process.operation.Operation):
    """Pull the project and records"""

    type = OperationsType.format

    def __init__(self, *, review_manager: colrev.review_manager.ReviewManager) -> None:
        super().__init__(
            review_manager=review_manager,
            operations_type=self.type,
        )

    @colrev.process.operation.Operation.decorate()
    def main(self) -> None:
        """Pull the CoLRev project and records (main entrypoint)"""

        self._pull_project()

    def _pull_project(self) -> None:
        try:
            git_repo = self.review_manager.dataset.get_repo()
            origin = git_repo.remotes.origin
            self.review_manager.logger.info(
                f"Pull project changes from {git_repo.remotes.origin}"
            )
            res = origin.pull()
        except AttributeError:
            self.review_manager.logger.info(
                f"{Colors.RED}No remote detected for pull{Colors.END}"
            )
            return

        if 4 == res[0].flags:
            self.review_manager.logger.info(
                f"{Colors.GREEN}Project up-to-date{Colors.END}"
            )
        elif 64 == res[0].flags:
            self.review_manager.logger.info(
                f"{Colors.GREEN}Updated CoLRev repository{Colors.END}"
            )
        else:
            self.review_manager.logger.info(
                f"{Colors.RED}Returned flag {res[0].flags}{Colors.END}"
            )
        print()



#! /usr/bin/env python
"""CoLRev pdf_get operation: Get PDF documents."""
from __future__ import annotations

import shutil
from glob import glob
from multiprocessing.pool import ThreadPool as Pool
from pathlib import Path

import colrev.exceptions as colrev_exceptions
import colrev.process.operation
import colrev.record.record_pdf
from colrev.constants import Colors
from colrev.constants import EndpointType
from colrev.constants import Fields
from colrev.constants import OperationsType
from colrev.constants import PDFPathType
from colrev.constants import RecordState
from colrev.writer.write_utils import write_file


class PDFGet(colrev.process.operation.Operation):
    """Get the PDFs"""

    to_retrieve: int
    retrieved: int
    not_retrieved: int

    type = OperationsType.pdf_get

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
        notify_state_transition_operation: bool = True,
    ) -> None:
        super().__init__(
            review_manager=review_manager,
            operations_type=self.type,
            notify_state_transition_operation=notify_state_transition_operation,
        )

        self.package_manager = self.review_manager.get_package_manager()

        pdf_dir = self.review_manager.paths.pdf
        pdf_dir.mkdir(exist_ok=True, parents=True)

        self.pdf_qm = self.review_manager.get_pdf_qm()

        self.filepath_directory_pattern = ""
        pdf_endpoints = [
            s
            for s in self.review_manager.settings.sources
            if s.endpoint == "colrev.files_dir"
        ]
        if pdf_endpoints:
            self.filepath_directory_pattern = (
                pdf_endpoints[0].search_parameters["scope"].get("subdir_pattern", {})
            )

    def copy_pdfs_to_repo(self) -> None:
        """Copy the PDFs to the repository"""
        self.review_manager.logger.info("Copy PDFs to dir")
        records = self.review_manager.dataset.load_records_dict()

        for record_dict in records.values():
            if Fields.FILE not in record_dict:
                continue
            fpath = Path(record_dict[Fields.FILE])
            new_fpath = fpath.absolute()
            if fpath.is_symlink():
                linked_file = fpath.resolve()
                if linked_file.is_file():
                    fpath.unlink()
                    shutil.copyfile(linked_file, new_fpath)
                    self.review_manager.logger.info(
                        f" {Colors.GREEN}copied PDF for {record_dict[Fields.ID]} {Colors.END}"
                    )

            elif new_fpath.is_file() and self.review_manager.verbose_mode:
                self.review_manager.logger.info(
                    f"No need to copy PDF - already exits ({record_dict[Fields.ID]})"
                )

    def link_pdf(
        self, record: colrev.record.record.Record
    ) -> colrev.record.record.Record:
        """Link the PDF in its record (should be {ID}.pdf)"""

        pdf_filepath = self.review_manager.paths.PDF_DIR / Path(
            f"{record.data['ID']}.pdf"
        )
        if (self.review_manager.path / pdf_filepath).is_file() and str(
            pdf_filepath
        ) != record.data.get(Fields.FILE, "NA"):
            record.update_field(
                key=Fields.FILE, value=str(pdf_filepath), source="link_pdf"
            )
            self.import_pdf(record)
            if RecordState.rev_prescreen_included == record.data[Fields.STATUS]:
                record.set_status(RecordState.pdf_imported)

        return record

    def get_target_filepath(self, record: colrev.record.record.Record) -> Path:
        """Get the target filepath for a PDF"""

        target_filepath = self.review_manager.paths.PDF_DIR / Path(
            f"{record.data['ID']}.pdf"
        )
        if self.filepath_directory_pattern == Fields.YEAR:
            target_filepath = self.review_manager.paths.PDF_DIR / Path(
                f"{record.data.get('year', 'no_year')}/{record.data['ID']}.pdf"
            )

        elif self.filepath_directory_pattern == "volume_number":
            if Fields.VOLUME in record.data and Fields.NUMBER in record.data:
                target_filepath = self.review_manager.paths.PDF_DIR / Path(
                    f"{record.data['volume']}/{record.data['number']}/{record.data['ID']}.pdf"
                )

            if Fields.VOLUME in record.data and Fields.NUMBER not in record.data:
                target_filepath = self.review_manager.paths.PDF_DIR / Path(
                    f"{record.data['volume']}/{record.data['ID']}.pdf"
                )

        return target_filepath

    def import_pdf(self, record: colrev.record.record.Record) -> None:
        """Import a file (PDF) and copy/symlink it"""
        # self.review_manager.pdf_dir.mkdir(exist_ok=True)
        # new_fp = self.review_manager.PDF_DIR_RELATIVE / Path(record.data[Fields.ID] + ".pdf").name
        new_fp = self.get_target_filepath(record)
        original_fp = Path(record.data[Fields.FILE])

        if new_fp != original_fp and not new_fp.is_file():
            new_fp.parents[0].mkdir(exist_ok=True, parents=True)
            if (
                PDFPathType.symlink
                == self.review_manager.settings.pdf_get.pdf_path_type
            ):
                new_fp.symlink_to(original_fp)
            elif PDFPathType.copy == self.review_manager.settings.pdf_get.pdf_path_type:
                shutil.copyfile(original_fp, new_fp.resolve())
            # Note : else: leave absolute paths

        record.data[Fields.FILE] = str(new_fp)

    def _log_infos(self, record: colrev.record.record.Record) -> None:
        if Fields.FILE not in record.data:
            if (
                not self.review_manager.settings.pdf_get.pdf_required_for_screen_and_synthesis
            ):
                return
            self.review_manager.logger.info(
                f" {Colors.ORANGE}{record.data['ID']}".ljust(46)
                + f"rev_prescreen_included → pdf_needs_manual_retrieval{Colors.END}"
            )
            return

        if RecordState.pdf_prepared == record.data[Fields.STATUS]:
            self.review_manager.logger.info(
                f" {Colors.GREEN}{record.data['ID']}".ljust(46)
                + f"rev_prescreen_included → pdf_prepared{Colors.END}"
            )
        elif RecordState.pdf_needs_manual_preparation == record.data[Fields.STATUS]:
            self.review_manager.logger.info(
                f" {Colors.ORANGE}{record.data['ID']}".ljust(46)
                + f"rev_prescreen_included → pdf_needs_manual_preparation{Colors.END}"
            )

    # Note : no named arguments (multiprocessing)
    def get_pdf(self, item: dict) -> dict:
        """Get PDFs (based on the package endpoints in the settings)"""

        record_dict = item["record"]

        if record_dict[Fields.STATUS] not in [
            RecordState.rev_prescreen_included,
            RecordState.pdf_needs_manual_retrieval,
        ]:
            if Fields.FILE in record_dict:
                record = colrev.record.record_pdf.PDFRecord(
                    record_dict, path=self.review_manager.path
                )
                record.remove_field(key=Fields.FILE)
                return record.get_data()
            return record_dict

        record = colrev.record.record_pdf.PDFRecord(
            record_dict, path=self.review_manager.path
        )

        for (
            pdf_get_package_endpoint
        ) in self.review_manager.settings.pdf_get.pdf_get_package_endpoints:

            pdf_get_class = self.package_manager.get_package_endpoint_class(
                package_type=EndpointType.pdf_get,
                package_identifier=pdf_get_package_endpoint["endpoint"],
            )
            endpoint = pdf_get_class(
                pdf_get_operation=self, settings=pdf_get_package_endpoint
            )

            endpoint.get_pdf(record)  # type: ignore

            if Fields.FILE in record.data:
                self.review_manager.report_logger.info(
                    f"{endpoint.settings.endpoint}"  # type: ignore
                    f"({record.data[Fields.ID]}): retrieved .../"
                    f"{Path(record.data[Fields.FILE]).name}"
                )
                break

        if Fields.FILE in record.data:
            record.run_pdf_quality_model(self.pdf_qm, set_prepared=True)
        else:
            record.set_status(RecordState.pdf_needs_manual_retrieval)

        self._log_infos(record)

        return record.get_data()

    def _fix_broken_symlinks(self) -> None:

        pdf_dir = self.review_manager.paths.pdf
        broken_symlinks = []
        for pdf_candidate in list(pdf_dir.glob("**/*.pdf")):
            relative_path = pdf_candidate.relative_to(self.review_manager.path)
            if not relative_path.is_file():
                if pdf_candidate.is_symlink():
                    broken_symlinks.append(pdf_candidate)
                continue

        if broken_symlinks:
            print("To fix broken symlinks:")
            old_path = input("Enter the old path: ")
            new_path = input("Enter the new path: ")
            for broken_symlink in broken_symlinks:
                new_file = str(broken_symlink.resolve()).replace(old_path, new_path)
                print(f"Fix {broken_symlink}")
                broken_symlink.unlink()
                broken_symlink.symlink_to(new_file)

    def _relink_pdfs_in_source(self, source: colrev.settings.SearchSource) -> None:

        # pylint: disable=too-many-locals

        self.review_manager.logger.info(
            "Checking PDFs in same directory to reassign when "
            f"the cpid is identical {source.filename}"
        )

        pdf_dir = self.review_manager.paths.pdf
        pdf_candidates = {}
        for pdf_candidate in list(pdf_dir.glob("**/*.pdf")):
            colrev_pdf_id = colrev.record.record_pdf.PDFRecord.get_colrev_pdf_id(
                pdf_candidate
            )
            relative_path = pdf_candidate.relative_to(self.review_manager.path)
            pdf_candidates[relative_path] = colrev_pdf_id

        source_records_dict = colrev.loader.load_utils.load(
            filename=source.filename,
            logger=self.review_manager.logger,
        )
        source_records = list(source_records_dict.values())
        corresponding_origin = str(source.filename)
        records = self.review_manager.dataset.load_records_dict()
        for record in records.values():
            if Fields.FILE not in record:
                continue

            # Note: we check the source_records based on the cpids
            # in the record because cpids are not stored in the source_record
            # (pdf hashes may change after import/preparation)
            source_rec = {}
            if corresponding_origin != "":
                source_origin_l = [
                    o for o in record[Fields.ORIGIN] if corresponding_origin in o
                ]
                if len(source_origin_l) == 1:
                    source_origin = source_origin_l[0]
                    source_origin = source_origin.replace(
                        f"{corresponding_origin}/", ""
                    )
                    source_rec_l = [
                        s for s in source_records if s[Fields.ID] == source_origin
                    ]
                    if len(source_rec_l) == 1:
                        source_rec = source_rec_l[0]

            if source_rec:
                if (
                    self.review_manager.path / Path(record[Fields.FILE])
                ).is_file() and (
                    self.review_manager.path / Path(source_rec[Fields.FILE])
                ).is_file():
                    continue
            else:
                if (self.review_manager.path / Path(record[Fields.FILE])).is_file():
                    continue

            self.review_manager.logger.info(record[Fields.ID])

            for pdf_candidate, cpid in pdf_candidates.items():
                if record.get("colrev_pdf_id", "") == cpid:
                    record[Fields.FILE] = str(pdf_candidate)
                    source_rec[Fields.FILE] = str(pdf_candidate)

                    self.review_manager.logger.info(
                        f"Found and linked PDF: {pdf_candidate}"
                    )
                    break

        if len(source_records) > 0:
            source_records_dict = {r[Fields.ID]: r for r in source_records}
            write_file(records_dict=source_records_dict, filename=source.filename)

        self.review_manager.dataset.save_records_dict(records)
        self.review_manager.dataset.add_changes(source.filename)

    def relink_pdfs(self) -> None:
        """Relink record files to the corresponding PDFs (if available)"""

        self._fix_broken_symlinks()

        sources = [
            s
            for s in self.review_manager.settings.sources
            if s.endpoint == "colrev.files_dir" and s.filename.is_file()
        ]
        for source in sources:
            self._relink_pdfs_in_source(source)

        self.review_manager.dataset.create_commit(msg="Relink PDFs")

    def check_existing_unlinked_pdfs(
        self,
        records: dict,
    ) -> dict:
        """Check for PDFs that are in the pdfs directory but not linked in the record file"""

        linked_pdfs = [
            str(Path(x[Fields.FILE]).resolve())
            for x in records.values()
            if Fields.FILE in x
        ]
        pdf_dir = self.review_manager.paths.pdf
        pdf_files = glob(str(pdf_dir) + "/**.pdf", recursive=True)
        unlinked_pdfs = [
            Path(x)
            for x in pdf_files
            if str(Path(x).resolve()) not in linked_pdfs
            and not any(kw in x for kw in ["_with_lp.pdf", "_with_cp.pdf", "_ocr.pdf"])
        ]

        if len(unlinked_pdfs) == 0:
            return records

        grobid_service = self.review_manager.get_grobid_service()
        grobid_service.start()
        self.review_manager.logger.info("Check unlinked PDFs")
        for file in unlinked_pdfs:
            msg = f"Check unlinked PDF: {file.relative_to(self.review_manager.path)}"
            self.review_manager.logger.info(msg)
            if file.stem not in records.keys():
                tei = self.review_manager.get_tei(pdf_path=file)
                pdf_record = tei.get_metadata()

                if "error" in pdf_record:
                    continue

                max_similarity = 0.0
                max_sim_record = None
                for record in records.values():
                    sim = colrev.record.record_pdf.PDFRecord.get_record_similarity(
                        colrev.record.record_pdf.PDFRecord(
                            pdf_record, path=self.review_manager.path
                        ),
                        colrev.record.record_pdf.PDFRecord(
                            record.copy(), path=self.review_manager.path
                        ),
                    )
                    if sim > max_similarity:
                        max_similarity = sim
                        max_sim_record = record
                if max_sim_record:
                    if max_similarity > 0.5:
                        if RecordState.pdf_prepared == max_sim_record[Fields.STATUS]:
                            continue

                        record = colrev.record.record_pdf.PDFRecord(
                            max_sim_record, path=self.review_manager.path
                        )
                        record.update_field(
                            key=Fields.FILE,
                            value=str(file),
                            source="linking-available-files",
                        )
                        self.import_pdf(record)
                        if (
                            RecordState.rev_prescreen_included
                            == record.data[Fields.STATUS]
                        ):
                            record.set_status(RecordState.pdf_imported)

                        self.review_manager.report_logger.info(
                            "linked unlinked pdf:" f" {file.name}"
                        )
                        self.review_manager.logger.info(
                            "linked unlinked pdf:" f" {file.name}"
                        )
                        # max_sim_record = \
                        #     pdf_prep.validate_pdf_metadata(max_sim_record)
                        # colrev_status = max_sim_record['colrev_status']
                        # if RecordState.pdf_needs_manual_preparation == colrev_status:
                        #     # revert?
            else:
                record = records[file.stem]
                self.link_pdf(
                    colrev.record.record_pdf.PDFRecord(
                        record, path=self.review_manager.path
                    )
                )

        self.review_manager.dataset.save_records_dict(records)

        return records

    def _rename_pdf(
        self,
        *,
        record_dict: dict,
        file: Path,
        new_filename: Path,
        pdfs_search_file: Path,
    ) -> None:
        record_dict[Fields.FILE] = new_filename
        if pdfs_search_file.is_file():
            colrev.env.utils.inplace_change(
                filename=pdfs_search_file,
                old_string="{" + str(file) + "}",
                new_string="{" + str(new_filename) + "}",
            )

        if not file.is_file():
            corrected_path = Path(str(file).replace("  ", " "))
            if corrected_path.is_file():
                file = corrected_path

        if file.is_file():
            shutil.move(str(file), str(new_filename))
        elif file.is_symlink():
            shutil.move(str(file), str(new_filename))

        record_dict[Fields.FILE] = str(new_filename)
        self.review_manager.logger.info(f"rename {file.name} > {new_filename}")
        if RecordState.rev_prescreen_included == record_dict[Fields.STATUS]:
            record = colrev.record.record_pdf.PDFRecord(
                record_dict, path=self.review_manager.path
            )
            record.set_status(RecordState.pdf_imported)

    def rename_pdfs(self) -> None:
        """Rename the PDFs"""

        self.review_manager.logger.info("Rename PDFs")

        records = self.review_manager.dataset.load_records_dict()

        # We may use other pdfs_search_files from the sources:
        # review_manager.settings.sources
        pdfs_search_file = Path("data/search/pdfs.bib")

        for record_dict in records.values():
            if Fields.FILE not in record_dict:
                continue
            if record_dict[Fields.STATUS] not in RecordState.get_post_x_states(
                state=RecordState.md_processed
            ):
                continue

            file = Path(record_dict[Fields.FILE])
            new_filename = file.parents[0] / Path(f"{record_dict['ID']}{file.suffix}")
            # Possible option: move to top (pdfs) directory:
            # new_filename = self.review_manager.PDF_DIR_RELATIVE / Path(
            #     f"{record['ID']}.pdf"
            # )
            if str(file) == str(new_filename):
                continue

            self._rename_pdf(
                record_dict=record_dict,
                file=file,
                new_filename=new_filename,
                pdfs_search_file=pdfs_search_file,
            )

        self.review_manager.dataset.save_records_dict(records)

        if pdfs_search_file.is_file():
            self.review_manager.dataset.add_changes(pdfs_search_file)

    def _get_data(self) -> dict:
        # pylint: disable=duplicate-code

        records_headers = self.review_manager.dataset.load_records_dict(
            header_only=True
        )
        record_header_list = list(records_headers.values())

        nr_tasks = len(
            [
                x
                for x in record_header_list
                if x[Fields.STATUS]
                in [
                    RecordState.pdf_needs_manual_retrieval,
                    RecordState.rev_prescreen_included,
                ]
            ]
        )

        items = self.review_manager.dataset.read_next_record(
            conditions=[
                {Fields.STATUS: RecordState.rev_prescreen_included},
                {Fields.STATUS: RecordState.pdf_needs_manual_retrieval},
            ],
        )

        self.to_retrieve = nr_tasks

        pdf_get_data = {
            "nr_tasks": nr_tasks,
            "items": [{"record": item} for item in items],
        }

        return pdf_get_data

    def _print_stats(self, retrieved_record_list: list) -> None:
        self.retrieved = len([r for r in retrieved_record_list if Fields.FILE in r])

        self.not_retrieved = self.to_retrieve - self.retrieved

        retrieved_string = "Overall pdf_imported".ljust(34)
        if self.retrieved == 0:
            retrieved_string += f"{self.retrieved}".rjust(6, " ")
            retrieved_string += " PDFs"
        elif self.retrieved == 1:
            retrieved_string += f"{Colors.GREEN}"
            retrieved_string += f"{self.retrieved}".rjust(6, " ")
            retrieved_string += f"{Colors.END} PDF"
        else:
            retrieved_string += f"{Colors.GREEN}"
            retrieved_string += f"{self.retrieved}".rjust(6, " ")
            retrieved_string += f"{Colors.END} PDFs"

        not_retrieved_string = "Overall pdf_needs_manual_retrieval".ljust(34)
        if self.not_retrieved == 0:
            not_retrieved_string += f"{self.not_retrieved}".rjust(6, " ")
            not_retrieved_string += " PDFs"
        elif self.not_retrieved == 1:
            not_retrieved_string += f"{Colors.ORANGE}"
            not_retrieved_string += f"{self.not_retrieved}".rjust(6, " ")
            not_retrieved_string += f"{Colors.END} PDF"
        else:
            not_retrieved_string += f"{Colors.ORANGE}"
            not_retrieved_string += f"{self.not_retrieved}".rjust(6, " ")
            not_retrieved_string += f"{Colors.END} PDFs"

        self.review_manager.logger.info(retrieved_string)
        self.review_manager.logger.info(not_retrieved_string)

    def _set_status_if_pdf_linked(self, records: dict) -> dict:
        for record_dict in records.values():
            if record_dict[Fields.STATUS] in [
                RecordState.rev_prescreen_included,
                RecordState.pdf_needs_manual_retrieval,
            ]:
                record = colrev.record.record_pdf.PDFRecord(
                    record_dict, path=self.review_manager.path
                )
                if Fields.FILE in record_dict:
                    if any(
                        Path(fpath).is_file()
                        for fpath in record.data[Fields.FILE].split(";")
                    ):
                        if (
                            RecordState.rev_prescreen_included
                            == record.data[Fields.STATUS]
                        ):
                            record.set_status(RecordState.pdf_imported)
                    else:
                        self.review_manager.logger.warning(
                            "Remove non-existent file link "
                            f"({record_dict[Fields.ID]}: {record_dict[Fields.FILE]}"
                        )
                        record.remove_field(key=Fields.FILE)
        self.review_manager.dataset.save_records_dict(records)

        return records

    def setup_custom_script(self) -> None:
        """Setup a custom pfd-get script"""

        filedata = colrev.env.utils.get_package_file_content(
            module="colrev.ops",
            filename=Path("custom_scripts/custom_pdf_get_script.py"),
        )
        if filedata:
            with open("custom_pdf_get_script.py", "w", encoding="utf-8") as file:
                file.write(filedata.decode("utf-8"))

        self.review_manager.dataset.add_changes(Path("custom_pdf_get_script.py"))

        self.review_manager.settings.pdf_get.pdf_get_man_package_endpoints.append(
            {"endpoint": "custom_pdf_get_script"}
        )

        self.review_manager.save_settings()

    @colrev.process.operation.Operation.decorate()
    def main(self) -> None:
        """Get PDFs (main entrypoint)"""

        if (
            self.review_manager.in_ci_environment()
            and not self.review_manager.in_test_environment()
        ):
            raise colrev_exceptions.ServiceNotAvailableException(
                dep="colrev pdf-prep",
                detailed_trace="pdf-prep not available in ci environment",
            )

        self.review_manager.logger.info("Get PDFs")
        self.review_manager.logger.info(
            "Get PDFs of prescreen-included records from local and remote sources."
        )
        self.review_manager.logger.info(
            "PDFs are stored in the directory data/pdfs "
            f"({Colors.ORANGE}colrev pdfs --dir{Colors.END})"
        )
        self.review_manager.logger.info(
            "See https://colrev-environment.github.io/colrev/manual/pdf_retrieval/pdf_get.html"
        )

        records = self.review_manager.dataset.load_records_dict()
        records = self._set_status_if_pdf_linked(records)
        records = self.check_existing_unlinked_pdfs(records)

        pdf_get_data = self._get_data()

        if pdf_get_data["nr_tasks"] == 0:
            self.review_manager.logger.info("No additional pdfs to retrieve")
        else:
            self.review_manager.logger.info(
                "PDFs to get".ljust(38) + f'{pdf_get_data["nr_tasks"]} PDFs'
            )

            pool = Pool(4)
            retrieved_record_list = pool.map(self.get_pdf, pdf_get_data["items"])
            pool.close()
            pool.join()

            self.review_manager.dataset.save_records_dict(
                {r[Fields.ID]: r for r in retrieved_record_list}, partial=True
            )

            self._print_stats(retrieved_record_list)

        # Note: rename should be after copy.
        # Note : do not pass records as an argument.
        if self.review_manager.settings.pdf_get.rename_pdfs:
            self.rename_pdfs()

        self.review_manager.dataset.create_commit(msg="PDFs: get and prepare")
        self.review_manager.logger.info(
            f"{Colors.GREEN}Completed pdf-get operation{Colors.END}"
        )



#! /usr/bin/env python
"""CoLRev pdf_prep_man operation: Prepare PDF documents manually."""
from __future__ import annotations

from pathlib import Path

import pandas as pd
import pymupdf

import colrev.exceptions as colrev_exceptions
import colrev.process.operation
import colrev.record.record
import colrev.record.record_pdf
from colrev.constants import EndpointType
from colrev.constants import Fields
from colrev.constants import Filepaths
from colrev.constants import OperationsType
from colrev.constants import RecordState
from colrev.writer.write_utils import write_file


class PDFPrepMan(colrev.process.operation.Operation):
    """Prepare PDFs manually"""

    type = OperationsType.pdf_prep_man

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
        notify_state_transition_operation: bool = True,
    ) -> None:
        super().__init__(
            review_manager=review_manager,
            operations_type=self.type,
            notify_state_transition_operation=notify_state_transition_operation,
        )

        self.verbose = True

    def discard(self) -> None:
        """Discard records whose PDFs need manual preparation (set to pdf_not_available)"""

        records = self.review_manager.dataset.load_records_dict()
        for record_dict in records.values():
            if record_dict[Fields.STATUS] == RecordState.pdf_needs_manual_preparation:
                record = colrev.record.record.Record(record_dict)
                record.set_status(RecordState.pdf_not_available)
        self.review_manager.dataset.save_records_dict(records)
        self.review_manager.dataset.create_commit(
            msg="Discard man-prep PDFs", manual_author=True
        )

    def get_data(self) -> dict:
        """Get the data for PDF prep man"""
        # pylint: disable=duplicate-code

        records_headers = self.review_manager.dataset.load_records_dict(
            header_only=True
        )
        record_header_list = list(records_headers.values())
        nr_tasks = len(
            [
                x
                for x in record_header_list
                if RecordState.pdf_needs_manual_preparation == x[Fields.STATUS]
            ]
        )
        pad = 0
        if record_header_list:
            pad = min((max(len(x[Fields.ID]) for x in record_header_list) + 2), 40)

        items = self.review_manager.dataset.read_next_record(
            conditions=[{Fields.STATUS: RecordState.pdf_needs_manual_preparation}]
        )
        pdf_prep_man_data = {"nr_tasks": nr_tasks, "PAD": pad, "items": items}
        self.review_manager.logger.debug(
            self.review_manager.p_printer.pformat(pdf_prep_man_data)
        )
        return pdf_prep_man_data

    def pdfs_prepared_manually(self) -> bool:
        """Check whether PDFs were prepared manually"""
        return self.review_manager.dataset.has_record_changes()

    def pdf_prep_man_stats(self) -> None:
        """Determine PDF prep man statistics"""
        # pylint: disable=duplicate-code

        self.review_manager.logger.info(
            f"Load {self.review_manager.paths.RECORDS_FILE}"
        )
        records = self.review_manager.dataset.load_records_dict()

        self.review_manager.logger.info("Calculate statistics")
        stats: dict = {Fields.ENTRYTYPE: {}}

        prep_man_hints = []
        crosstab = []
        for record_dict in records.values():
            if RecordState.pdf_needs_manual_preparation != record_dict[Fields.STATUS]:
                continue

            if record_dict[Fields.ENTRYTYPE] in stats[Fields.ENTRYTYPE]:
                stats[Fields.ENTRYTYPE][record_dict[Fields.ENTRYTYPE]] = (
                    stats[Fields.ENTRYTYPE][record_dict[Fields.ENTRYTYPE]] + 1
                )
            else:
                stats[Fields.ENTRYTYPE][record_dict[Fields.ENTRYTYPE]] = 1

            record = colrev.record.record.Record(record_dict)
            prov_d = record.data[Fields.D_PROV]

            if Fields.FILE in prov_d:
                if prov_d[Fields.FILE]["note"] != "":
                    for hint in prov_d[Fields.FILE]["note"].split(","):
                        prep_man_hints.append(hint.lstrip())

            for hint in prep_man_hints:
                crosstab.append([record_dict[Fields.JOURNAL], hint.lstrip()])

        crosstab_df = pd.DataFrame(crosstab, columns=[Fields.JOURNAL, "hint"])

        if crosstab_df.empty:
            print("No records to prepare manually.")
        else:
            # pylint: disable=duplicate-code
            tabulated = pd.pivot_table(
                crosstab_df[[Fields.JOURNAL, "hint"]],
                index=[Fields.JOURNAL],
                columns=["hint"],
                aggfunc=len,
                fill_value=0,
                margins=True,
            )
            # .sort_index(axis='columns')
            tabulated.sort_values(by=["All"], ascending=False, inplace=True)
            # Transpose because we tend to have more error categories than search files.
            tabulated = tabulated.transpose()
            print(tabulated)
            self.review_manager.logger.info(
                "Writing data to file: manual_preparation_statistics.csv"
            )
            tabulated.to_csv("manual_pdf_preparation_statistics.csv")

    def extract_needs_pdf_prep_man(self) -> None:
        """Apply PDF prep man to csv/bib"""

        prep_bib_path = self.review_manager.path / Path("data/pdf-prep-records.bib")
        prep_csv_path = self.review_manager.path / Path("data/pdf-prep-records.csv")

        if prep_csv_path.is_file():
            print(f"Please rename file to avoid overwriting changes ({prep_csv_path})")
            return

        if prep_bib_path.is_file():
            print(f"Please rename file to avoid overwriting changes ({prep_bib_path})")
            return

        self.review_manager.logger.info(
            f"Load {self.review_manager.paths.RECORDS_FILE}"
        )
        records = self.review_manager.dataset.load_records_dict()

        records = {
            record[Fields.ID]: record
            for record in records.values()
            if RecordState.pdf_needs_manual_preparation == record[Fields.STATUS]
        }

        write_file(records_dict=records, filename=prep_bib_path)

        bib_db_df = pd.DataFrame.from_records(list(records.values()))

        # pylint: disable=duplicate-code
        col_names = [
            Fields.ID,
            Fields.ORIGIN,
            Fields.AUTHOR,
            Fields.TITLE,
            Fields.YEAR,
            Fields.JOURNAL,
            # Fields.BOOKTITLE,
            Fields.VOLUME,
            Fields.NUMBER,
            Fields.PAGES,
            Fields.DOI,
        ]
        for col_name in col_names:
            if col_name not in bib_db_df:
                bib_db_df[col_name] = "NA"
        bib_db_df = bib_db_df[col_names]

        bib_db_df.to_csv(prep_csv_path, index=False)
        self.review_manager.logger.info(f"Created {prep_csv_path.name}")

    def apply_pdf_prep_man(self) -> None:
        """Apply PDF prep man from csv/bib"""

        if Path("data/pdf-prep-records.csv").is_file():
            self.review_manager.logger.info("Load prep-records.csv")
            bib_db_df = pd.read_csv("data/pdf-prep-records.csv")
            records_changed = bib_db_df.to_dict("records")

        if Path("data/pdf-prep-records.bib").is_file():
            self.review_manager.logger.info("Load prep-records.bib")

            records_changed = colrev.loader.load_utils.load(
                filename=Path("data/pdf-prep-records.bib"),
                logger=self.review_manager.logger,
            )

        records = self.review_manager.dataset.load_records_dict()
        for record in records.values():
            # IDs may change - matching based on origins
            changed_record_l = [
                x for x in records_changed if x[Fields.ORIGIN] == record[Fields.ORIGIN]
            ]
            if len(changed_record_l) == 1:
                changed_record = changed_record_l.pop()
                for key, value in changed_record.items():
                    # if record['ID'] == 'Alter2014':
                    #     print(key, value)
                    if str(value) == "nan":
                        if key in record:
                            del record[key]
                        continue
                    record[key] = value
                    if value == "":
                        del record[key]

        self.review_manager.dataset.save_records_dict(records)
        self.review_manager.check_repo()

    def extract_coverpage(self, *, filepath: Path) -> None:
        """Extract coverpage from PDF"""

        Filepaths.COVERPAGES.mkdir(exist_ok=True)

        doc1 = pymupdf.Document(str(filepath))
        if doc1.page_count > 0:
            colrev.record.record_pdf.PDFRecord.extract_pages_from_pdf(
                pages=[0],
                pdf_path=filepath,
                save_to_path=Filepaths.COVERPAGES,
            )

    def extract_lastpage(self, *, filepath: Path) -> None:
        """Extract last page from PDF"""

        Filepaths.LASTPAGES.mkdir(exist_ok=True)

        doc1 = pymupdf.Document(str(filepath))
        if doc1.page_count > 0:
            colrev.record.record_pdf.PDFRecord.extract_pages_from_pdf(
                pages=[doc1.page_count - 1],
                pdf_path=filepath,
                save_to_path=Filepaths.LASTPAGES,
            )

    def extract_pages(self, *, filepath: Path, pages_to_remove: list) -> None:
        """Extract pages from PDF"""

        doc1 = pymupdf.Document(str(filepath))
        if doc1.page_count > 0:
            colrev.record.record_pdf.PDFRecord.extract_pages_from_pdf(
                pages=pages_to_remove,
                pdf_path=filepath,
            )

    def set_pdf_man_prepared(self, record: colrev.record.record.Record) -> None:
        """Set the PDF to manually prepared"""

        record.set_status(RecordState.pdf_prepared)
        record.reset_pdf_provenance_notes()

        pdf_path = Path(self.review_manager.path / Path(record.data[Fields.FILE]))
        prev_cpid = record.data.get("colrev_pdf_id", "NA")
        record.data.update(colrev_pdf_id=record.get_colrev_pdf_id(pdf_path))
        if prev_cpid != record.data.get("colrev_pdf_id", "NA"):
            record.add_field_provenance(key=Fields.FILE, source="manual")

        record_dict = record.get_data()
        self.review_manager.dataset.save_records_dict(
            {record_dict[Fields.ID]: record_dict}, partial=True
        )
        self.review_manager.dataset.add_changes(self.review_manager.paths.RECORDS_FILE)

    @colrev.process.operation.Operation.decorate()
    def main(self) -> None:
        """Prepare PDFs manually (main entrypoint)"""

        if (
            self.review_manager.in_ci_environment()
            and not self.review_manager.in_test_environment()
        ):
            raise colrev_exceptions.ServiceNotAvailableException(
                dep="colrev pdf-prep-man",
                detailed_trace="pdf-prep-man not available in ci environment",
            )

        records = self.review_manager.dataset.load_records_dict()

        package_manager = self.review_manager.get_package_manager()

        pdf_prep_man_package_endpoints = (
            self.review_manager.settings.pdf_prep.pdf_prep_man_package_endpoints
        )
        for pdf_prep_man_package_endpoint in pdf_prep_man_package_endpoints:

            pdf_prep_man_class = package_manager.get_package_endpoint_class(
                package_type=EndpointType.pdf_prep_man,
                package_identifier=pdf_prep_man_package_endpoint["endpoint"],
            )
            endpoint = pdf_prep_man_class(
                pdf_prep_man_operation=self,
                settings=pdf_prep_man_package_endpoint,
            )

            records = endpoint.pdf_prep_man(records)  # type: ignore
